/*
CRITSECT.C
Andrew Schulman, January 1993
CIS 76320,302
from "Undocumented DOS", 2nd edition (Addison-Wesley, 1993)

Thanks to John Brennan (John.Brennan@vi.ri.cmu.edu)
and to Norman D. Culver (CIS 70672,1257)

Reports on DOS critical sections, turns off or on

Table of critical section patch offsets at:
DOS:02C3h in DOS 3.1-3.3
DOS:0315h in DOS 4, 5, 6

to turn ON critical section calls in DOS:
in DOS 3-4, poke in 50h (PUSH AX) to replace C3h (RET)
in DOS 5-6, poke in any non-zero value
*/

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <dos.h>

#ifndef MK_FP
#define MK_FP(seg, ofs) \
    ((void far *) (((unsigned long) (seg) << 16) | (ofs)))
#endif      

// get DOS data segment
// call 21/52 (Get List of Lists), keep segment, discard offset
unsigned short get_dos_ds(void)
{
    _asm push es
    _asm xor ax, ax
    _asm mov es, ax
    _asm mov ax, 5200h
    _asm int 21h
    _asm mov ax, es
    _asm pop es
    // return value in AX
}

// get DOS version number
// if DOS 5+, use 21/3306 because of SETVER
void get_dos_vers(unsigned char *pmaj, unsigned char *pmin)
{
    unsigned char maj, min;
	// must call 21/3306 before 21/30
	// but don't know if 21/3306 supported!
	// best bet is to clear BX beforehand
	// unsupported should set AL=FFh, but not in DR-DOS?
	_asm xor bx, bx
	_asm mov ax, 3306h
	_asm int 21h
	_asm mov maj, bl
	_asm mov min, bh
	if ((maj==0) && (min==0))
	{
		_asm mov ax, 3000h
		_asm int 21h
		_asm mov maj, al
		_asm mov min, ah
	}
    *pmaj = maj;
    *pmin = min;
}

void fail(char *s) { puts(s); exit(1); }

typedef enum { STATUS, ON, OFF } REQU;

main(int argc, char *argv[])
{
    unsigned short far *patch_tab;
    unsigned short far *p;
    unsigned char far *fpop;
    unsigned short dos_ds;
    unsigned short tab_ofs;
    unsigned char dos_maj, dos_min;
    char *s;
    unsigned char op;
    REQU requ;
    
    // command line:  ON, OFF, or (default) STATUS
    if (argc < 2) 
        requ = STATUS;
    else
    {
        s = strupr(argv[1]);
        if (strcmp(s, "STATUS") == 0) requ = STATUS;
        else if (strcmp(s, "ON") == 0) requ = ON;
        else if (strcmp(s, "OFF") == 0) requ = OFF;
        else fail("usage: critsect [status | on | off]");
    }
        
    // get DOS data segment
    dos_ds = get_dos_ds();
        
    // get DOS version number -- program works with DOS 3.1-6.0
    // may have to check for DR-DOS, etc.!!
    get_dos_vers(&dos_maj, &dos_min);
    if ((dos_maj < 3) || (dos_maj > 6) || ((dos_maj == 3) && (dos_min == 0)))
        fail("Unsupported DOS version");
        
    // get pointer to table
    tab_ofs = (dos_maj == 3) ? 0x02c3 : 0x0315;
    patch_tab = (unsigned short far *) MK_FP(dos_ds, tab_ofs);
    
    if (requ == STATUS)
        printf("Critical section patch table at %Fp\n", patch_tab);
    
#define RET         0xC3
#define PUSH_AX     0x50

    // walk table
    for (p=patch_tab; *p != 0; p++)
    {
        fpop = (unsigned char far *) MK_FP(dos_ds, *p);
        op = *fpop;
        if (dos_maj < 5)	// IS THIS WRONG???  RALF
        {
            if (op == RET) s = "OFF";
            else if (op == PUSH_AX) s = "ON";
            else fail("Critical section patch table invalid!");
            
            if (requ == ON)         *fpop = PUSH_AX;
            else if (requ == OFF)   *fpop = RET;
        }
        else
        {
            if (op == 0) s = "OFF";
            else s = "ON";
            // no way to check for valid/invalid,
            // though anything other than 0 or 1 is suspicious
                
            if (requ == ON)         *fpop = 1;
            else if (requ == OFF)   *fpop = 0;
        }

        if (requ == STATUS)
            printf("%Fp = %02X  %s\n", fpop, op, s);
    }
    
    if (requ == ON) puts("Turned critical sections ON");
    else if (requ == OFF) puts("Turned critical sections OFF");
    
    return 0;
}
