/* 
DEV.C -- display MS-DOS device chain -- for Windows
Andrew Schulman, March 1993
From "Undocumented DOS", 2nd edition (Addison-Wesley, 1993)

Real mode DOS:
bcc dev.c

Windows:
bcc -c -2 -DWINDOWS -W prot.c printf.c
tlib win_app-+prot.c-+printf.c
bcc -2 -DWINDOWS -W dev.c win_app.lib
*/

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <dos.h>
#ifdef WINDOWS
#include <windows.h>
#include "prot.h"
#include "printf.h"
#endif

#ifdef WINDOWS
char *app = "Walk DOS Device Chain";
#define puts(s)         MessageBox(NULL, s, app, MB_OK)
#endif

// this fail() not suitable for a DPMI_APP (which would
// need to do a _dos_exit()
void fail(const char *s, ...) { puts(s); exit(1); }

#ifndef MK_FP
#define MK_FP(seg, ofs) \
    ((void far *) (((unsigned long) (seg) << 16) | (ofs)))
#endif      

#ifdef WINDOWS
#define MAP(ptr, bytes)     map_real((ptr), (bytes))
#define FREE_MAP(ptr)       free_mapped_linear(ptr)
#define GET_REAL(ptr)       get_real_addr(ptr)
#define YIELD()             Yield()
#else
#define MAP(ptr, bytes)     (ptr)
#define FREE_MAP(ptr)       /**/
#define GET_REAL(ptr)       (ptr)
#define YIELD()             /**/
#endif

/* some device attribute bits */
#define CHAR_DEV    (1 << 15)
#define INT29       (1 << 4)
#define IS_CLOCK    (1 << 3)
#define IS_NUL      (1 << 2)

#pragma pack(1)

typedef struct DeviceDriver {
    struct DeviceDriver far *next;
    unsigned attr;
    unsigned strategy;
    unsigned intr;
    union {
        unsigned char name[8];
        unsigned char blk_cnt;
        } u;
    } DeviceDriver;

typedef struct {
    unsigned char misc[8];
    DeviceDriver far *clock;
    DeviceDriver far *con;
    unsigned char misc2[18];
    DeviceDriver nul;   /* not a pointer */
    // ...
    } ListOfLists;  // DOS 3.1+
        
ListOfLists far *get_doslist(void)
{
    union REGS r;
    struct SREGS s;
    memset(&r, 0, sizeof(r));
    memset(&s, 0, sizeof(s));
    r.x.ax = 0x5200;
#ifdef WINDOWS  
    /* If Windows, call undocumented DOS INT 21h Function 52h via DPMI
       "Simulate Real Mode Interrupt" call (INT 31h AX=0300h), and return
       the resulting real mode pointer. */
    real_int86x(0x21, &r, &r, &s);
#else
    int86x(0x21, &r, &r, &s);
#endif
    return (ListOfLists far *) MK_FP(s.es, r.x.bx);
}

#ifdef WINDOWS
#ifdef __cplusplus
extern "C" int PASCAL WinMain(HANDLE hInstance, HANDLE hPrevInstance,
    LPSTR lpszCmdLine, int nCmdShow);
#endif

#pragma argsused
int PASCAL WinMain(HANDLE hInstance, HANDLE hPrevInstance,
    LPSTR lpszCmdLine, int nCmdShow)
#else       
int main(int argc, char *argv[])
#endif
{
    ListOfLists far *doslist;
    DeviceDriver far *dd;
    DeviceDriver far *next;
    
#ifdef WINDOWS
    if (! (GetWinFlags() & WF_PMODE))
        fail("This program requires Windows Standard or Enhanced mode");
    
    // could also do protected mode DOS version with DPMI
#endif  

    if (! (doslist = get_doslist()))
        fail("INT 21h Function 52h not supported");
    
#ifdef WINDOWS  
    /* get protected mode pointer to DOS internal variable table */
    doslist = (ListOfLists far *) map_real(doslist, sizeof(ListOfLists));
    open_display(app);
#endif  

#define PARANOID
#ifdef  PARANOID
    /* This block of code just double-checks that everything is ok */

    /* NUL is part of DOSLIST, not a pointer, so don't need to map */
    if (_fmemcmp(doslist->nul.u.name, "NUL     ", 8) != 0)
        fail("NUL name wrong");
    if (! (doslist->nul.attr & IS_NUL))
        fail("NUL attr wrong");
    
    /* CON is pointer, so need to map */
    dd = (DeviceDriver far *) MAP(doslist->con, sizeof(DeviceDriver));
    if (_fmemcmp(dd->u.name, "CON     ", 8) != 0)
        fail("CON name wrong");
    if (! (dd->attr & CHAR_DEV))
        fail("CON attr wrong");
    FREE_MAP(dd);
    
    /* CLOCK$ is also pointer, so need to map */
    dd = (DeviceDriver far *) MAP(doslist->clock, sizeof(DeviceDriver));
    if (_fmemcmp(dd->u.name, "CLOCK$  ", 8) != 0)
        fail("CLOCK$ name wrong");
    if (! (dd->attr & IS_CLOCK))
        fail("CLOCK$ attr wrong");
    FREE_MAP(dd);
#endif /*PARANOID*/

    /* 
       Print out device chain:  thanks to the MAP, FREE_MAP, GET_REAL,
       and YIELD macros, this works in real mode DOS or in protected mode
       Windows.  Old real mode only code looked like this:
            do { 
                printf("%Fp\t", dd);
                if (dd->attr & CHAR_DEV)
                    printf("%.8Fs\n", dd->u.name); 
                else
                    printf("Block dev: %u unit(s)\n", dd->u.blk_cnt);
                dd = dd->next;
            } while (FP_OFF(dd->next) != -1);
    */
    for (dd = &doslist->nul;;) 
    { 
        printf("%Fp    ", GET_REAL(dd)); /* print real mode addr */
        if (dd->attr & CHAR_DEV)
        {
#ifdef __BORLANDC__
            // Borland C++ printf can't handle non-terminated strings
            char buf[9];
            _fmemcpy(buf, dd->u.name, 8);
            buf[8] = '\0';
            printf("%s\r\n", buf);
#else
            printf("%.8Fs\r\n", dd->u.name);
#endif      
        }
        else
            printf("Block dev: %u unit(s)\r\n", dd->u.blk_cnt);
        next = dd->next;            /* get next pointer */
        /* first time through, this will free selector to doslist */
        FREE_MAP(dd);               /* THEN free rmode seg */
        if (FP_OFF(next) == 0xFFFF) /* is there a next? */
            break;
        dd = (DeviceDriver far *) MAP(next, sizeof(DeviceDriver));
        YIELD();    /* no message loop in this program, so yield */
    }
	
{
	int i;
	for (i=0; i<15; i++) printf("\n");
}
    
#ifdef WINDOWS  
{
    WORD mapped;
    if (mapped = get_mapped())
        printf("Error!: %u remaining mapped selectors!\r\n", mapped);
    show_display();
    return mapped;  /* 0 indicates success */
}
#else
    return 0;
#endif
}
