/*
FILETYPE.C 
Win32 Console app to clearly (?) display all registered file types

Andrew Schulman
andrew@ora.com
August 1995

cl filetype.c advapi32.lib

Sample output:
Word.Document.6 (".DOC")
    open "C:\WINWORD\WINWORD.EXE /w"
        '[FileOpen("%1")]'
    print "C:\WINWORD\WINWORD.EXE /w"
        '[FileOpen("%1")][FilePrint()][DocClose(2)]'
            
htmlfile (".html")
    open '"C:\Program Files\Plus!\Microsoft Internet\IExplore.exe" -nohome'
        '"file:%1",,-1,,,,,'

htmlfile (".htm")
    open '"C:\Program Files\Plus!\Microsoft Internet\IExplore.exe" -nohome'
        '"file:%1",,-1,,,,,'
            
msgfile (".msg")
    open "C:\Progra~1\Micros~1\exchng32.exe /f %1"
    print "C:\Progra~1\Micros~1\exchng32.exe /p %1"

etc.
*/

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#define WIN32_LEAN_AND_MEAN
#include "windows.h"            // includes winnt.h and winreg.h

#define HKCR        HKEY_CLASSES_ROOT
#define OK          ERROR_SUCCESS

int GET_VALUE(HKEY hkey, char *keystr, char *value)
{
    int cb = 256;       // sizeof(value) must be 256+
    value[0] = '\0';
    return ((RegQueryValue(hkey, keystr, value, &cb) == OK) && cb && *value);
}

void fail(const char *s) { puts(s); exit(1); }

main(int argc, char *argv[])
{
    HKEY hkey;
    char key[256], value[256], verb[256], menu[256];
	char defverb[256], cmd[256], key2[256];
    int i, j;

    // find all file-extension entries (start with '.')
    for (i = 0; 
         RegEnumKey(HKCR, i, key, 256) == OK;
         i++)
        if (key[0] == '.')
    {
        // get the value of this file-extension entry
        // example: .DOC -> "Word.Document.6"
        // NOTE: MIME types have no value
        if (GET_VALUE(HKCR, key, value))    
        {
            printf("%s (\"%s\")\n", value, key);
            
            // now treat that value as a key, and look for key\shell
            // example: Word.Document.6\shell
            strcat(value, "\\shell");
            if (RegOpenKey(HKCR, value, &hkey) == OK)
            {
				// see if there's a default action
				// if so, below will mark it with an '*'
				// "default default" action is open, so unmarked
				GET_VALUE(hkey, NULL, defverb);
					
                // enumerate all the verbs under key\shell
                // example: open, print, printto, explore
                for (j=0;
                     RegEnumKey(hkey, j, verb, 256) == OK;
                     j++)
                {
                    printf("\t%c%s ", 
						((strcmp(verb, defverb) == 0) ? '*' : ' '),
						verb);
                    
                    // get any menu string associated with this verb
                    if (GET_VALUE(hkey, verb, menu))
                        printf("(\'%s\') ", menu);
                    
                    // what command line is associated with this verb?
                    sprintf(key2, "%s\\%s\\command", value, verb);
                    if (GET_VALUE(HKCR, key2, cmd))
                    {
                        if (strchr(cmd, '\"'))  printf("\'%s\'\n", cmd);
                        else                    printf("\"%s\"\n", cmd);
                    }
					else						putchar('\n');
                        
                    // see if DDE exec string is associated with this verb
                    sprintf(key2, "%s\\%s\\ddeexec", value, verb);
                    if (GET_VALUE(HKCR, key2, cmd))
                    {
                        if (strchr(cmd, '\"'))  printf("\t\t\'%s\'\n", cmd);
                        else                    printf("\t\t\"%s\"\n", cmd);
                    }
                }
                RegCloseKey(hkey);
            }
            putchar('\n');
        }
    }
}
