/*
graph.c
Andrew Schulman
3 May 1989

Graph Theory subroutines for use by euler.c and spanning.c
*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <malloc.h>
#include "graph.h"

MATRIX read_graph(void)
{
    char buf[256];
    char ch;
    MATRIX m;
    WORD r,c;
    WORD nodes;
    char *s;
    
    /* skip blank lines */
    while ((ch = getchar()) == '\n')
        ;
    if (ch == EOF)
        return ((MATRIX) 0);
    ungetc(ch, stdin);

    nodes = atoi(gets(buf));
    m = matrix(nodes,nodes);
    r = 0;
    
    while (gets(buf))       /* get next row */
    {
        r++;
        if (r > nodes)
            break;
        if (s = strtok(buf, " \t,"))
        {
            int xnode, ynode;
            xnode = atoi(s);
            c = 1;
            while (s = strtok(NULL," \t,")) /* get next element within row */
            {
                c++;
                if (c > nodes)
                    break;
                ynode = atoi(s);
#ifdef MULTIGRAPH
                m[xnode][ynode]++;
                m[ynode][xnode]++;
#else   
                m[xnode][ynode] = m[ynode][xnode] = 1; /* store redundantly */
#endif          
            }
        }
        else
            break;
    }

    if (! r)
        fail("read_graph failure");
    
    return m;
}

void print_graph(MATRIX m)
{
    WORD i,j;
    if (! m) return;
    for (i=1; i<=ROWS(m); i++)
    {
        for (j=1; j<=COLS(m); j++)
            printf("%u ", m[i][j]);
        printf("\n");
    }
}

/**********************************************************************/
/* note that calloc sets all elements to 0 */
MATRIX matrix(WORD rows, WORD cols)
{
    WORD i;
    MATRIX m;
    
    /* allocate pointers to rows */
    m = (MATRIX) calloc(rows+1, sizeof(ROW));
    if (! m) 
        fail("matrix alloc failure #1");
    
    /* store header information */
    m--;
    HEADER(m) = calloc(2,2);
    ROWS(m) = rows;
    COLS(m) = cols;
    
    /* allocate rows and set pointers to them */
    for (i=1; i <= rows; i++)
    {
        m[i] = (ROW) calloc(cols+1, sizeof(WORD));
        if (! m[i]) 
            fail("matrix alloc failure #2");
        m[i]--;
    }
    
    return m;
}

void free_matrix(MATRIX m)
{
    WORD i;

    if (! m) return;
    free(HEADER(m));
    for (i=1; i<=ROWS(m); i++)
        free(m[i]);
    free(m);
}

void fail(char *msg)
{
    puts(msg);
    exit(1);
}

/**********************************************************************/
LIST list(WORD nodes)
{
    LIST l = ((LIST) calloc(nodes+1, sizeof(NODE)));
    if (! l) return (LIST) 0;
    SIZE(l) = nodes;
    return l;
}

void add(LIST l, WORD i, WORD info)
{
    if (! l) return;
    if (! l[i].info)                /* see if list for this node is empty */
    {
        l[i].info = info;
    }
    else
    {
        NODE *q;
        NODE *p = l+i;
        do
        {
#ifdef MULTIGRAPH
            /* allow duplicates for multigraphs */
#else
            if (p->info == info)    /* see if it's already there */
                return;
#endif          
            q = p;
        } while (p = p->ptr);
        
        /* at end of list; add the new info */
        q->ptr = calloc(1, sizeof(NODE));
        p = q->ptr;
        p->info = info;
    }
}

void print_list(LIST l)
{
    WORD i;

    if (! l) return;
    for (i=1; i<= SIZE(l); i++)
    {
        NODE *p = l+i;
        printf("[%u] ", i);
        do
        {
            if (p->info)
                printf("%u ", p->info);
        } while (p = p->ptr);
        printf("\n");
    }
}

LIST mat_to_list(MATRIX m)
{
    LIST l;
    WORD i,j, k;

    if (! m) return ((MATRIX) 0);
    l = list(ROWS(m));
    for (i=1; i<=ROWS(m); i++)
        for (j=1; j<=COLS(m); j++)
#ifdef MULTIGRAPH
            for (k=0; k<m[i][j]; k++)
                add(l,i,j);
#else
            if (m[i][j])
                add(l,i,j);
#endif          
    return l;
}

/**********************************************************************/
QUEUE queue(WORD size)
{
    QUEUE q = calloc(size+STARTQ, sizeof(WORD));
    HEAD(q) = STARTQ;
    TAIL(q) = STARTQ;
    SIZEQ(q) = size;
    return q;
}

void put(QUEUE q, WORD i)
{
    q[TAIL(q)] = i;
    TAIL(q)++;
}

WORD get(QUEUE q)
{
    WORD x;
    if (HEAD(q) == TAIL(q))
        fail("queue empty");
    x = q[HEAD(q)];
    HEAD(q)++;
    return x;
}

BOOL empty(QUEUE q)
{
    return (HEAD(q) == TAIL(q));
}

/* for debugging purposes */
void dump_queue(QUEUE q)
{
    WORD i;
    printf("Q: {");
    for (i=HEAD(q); i<TAIL(q); i++)
        printf("%d ", q[i]);
    printf("}\n");
}

void free_queue(QUEUE q)
{
    if (! q) return;
    if (! empty(q)) printf("Queue not empty!\n");
    if (q) free(q);
}

