/* 
HASHTAB.C
Andrew Schulman

To use:
    copy code
    change NODE structure to include your data instead of x,y
    change insert(), find(), print_list() to include your data instead of x,y
    change set(), get(), print_hashtab() to include your data instead of x,y
        
To do:
    create functions for *deleting* list and table items
    test list and table deletion
    
Most of the code here is for testing

usage:   hashtab [num items] [hash table size]
default: hashtab 2111 1021

Sample output (hashtab 20000 1021):
    Table size:       1021
    List insertions:  20000
    List finds:       2000000
    List find/whiles: 29321899
    Max whiles:       33
    Average whiles:   14
    Hash sets:        20000
    Hash gets:        2000000
    Not found:        1084264
    Clocks:           470
    Seconds:          25.824176
    Bytes allocated:  164084
    Items allocated:  20001
        
2111
----
                        No command switches         Optimize switches
                                            
Borland 16-bit  int         0.93 sec                    0.82 sec
Borland 16-bit -ml long     1.26                        1.15
Microsoft 32-bit            1.15                        0.87
        
21111
-----

Borland 16-bit int                  insufficient memory
Borland 16-bit -ml int      35.7                        31.8
Borland 16-bit -ml long     44.7                        41.6
Microsoft 32-bit            31.4                        25.1 (-G5), 26.0 (-G3)

Borland optimize switches = -Ox -3
Microsoft optimize switches = -Ox -Gr -G5 (or -G3)

Examine most frequently executed block of code:

        if (list->x == x)
            { if (whiles>max_whiles) max_whiles = whiles; return list->y; }
        else
            list = list->next;
        
Borland 16-bit:

    cmp word ptr [si+2],dx
    jne short @5@170
        ...
    mov si,word ptr [si]
@5@198:
    or  si,si
    jne short @5@58
        
Microsoft 32-bit:

    cmp DWORD PTR [ecx+4], edx
    je  SHORT $L466
    mov ecx, DWORD PTR [ecx]
    or  ecx, ecx
    jne SHORT $L388
        
So far, look pretty similar.
Now Borland 16-bit with -ml -DUINT=long:

    les bx,dword ptr [bp+6]
    mov ax,word ptr es:[bx+6]
    mov dx,word ptr es:[bx+4]
    cmp ax,word ptr [bp+12]
    jne short @5@254
    cmp dx,word ptr [bp+10]
    jne short @5@254
        ...
    les bx,dword ptr [bp+6]
    mov ax,word ptr es:[bx+2]
    mov dx,word ptr es:[bx]
    mov word ptr [bp+8],ax
    mov word ptr [bp+6],dx
@5@282:
    mov ax,word ptr [bp+6]
    or  ax,word ptr [bp+8]
    jne short @5@58
        
This is sort of thing that 32-bit code saves us from!!
But only when manipulating 32-bit quantities from 16-bit program.
Like, duh!

Oops, have to compare 16-bit protected mode code with 32-bit.
Would make different for large model pointer.
*/

#include <stdlib.h>
#include <stdio.h>
#include <time.h>

#ifndef UINT
#define UINT unsigned int
#endif

typedef struct _node {
    struct _node *next;
    UINT x;
    UINT y;
    } NODE;

// these are all for checking that searches aren't too lengthy  
UINT long inserts=0, finds=0, find_whiles=0, sets=0, getss=0;
UINT long not_found=0, mem_alloc=0, mem_items=0;
UINT whiles, max_whiles=0;

void fail(const char *s) { puts(s); exit(1); }

void *allocate(size_t x, size_t y)
{
    void *p;
    if (! x) x = 1;
    if (! y) y = 1;
    if (((long) x * (long) y) > 0xFFFFUL)
        fail("internal error");
    if ((p = calloc(x, y)) == 0)
        fail("insufficient memory");
    mem_alloc += (x * y);
    mem_items++;
    return p;
}
    
NODE *mk_list(void)
{
    return (NODE *) 0;
}
    
void insert(NODE **list, UINT x, UINT y)
{
    NODE *node = (NODE *) allocate(1, sizeof(NODE));
    inserts++;
    node->next = *list;
    *list = node;
    node->x = x;
    node->y = y;
}

UINT find(NODE *list, UINT x)
{
    finds++;
    whiles = 0;
    while (list)
    {
        whiles++;
        find_whiles++;
        if (list->x == x)
            { if (whiles>max_whiles) max_whiles = whiles; return list->y; }
        else
            list = list->next;
    }
    if (whiles>max_whiles) max_whiles = whiles;
    return 0;   // didn't find:  needs to be element not in set!
}

void print_list(NODE *list)
{
    while (list)
    {
        printf("[%u,%u] ", list->x, list->y);
        list = list->next;
    }
    printf("\n");
}

UINT TAB_SIZE = 1021;

#define HASH(x)         ((x) % TAB_SIZE)

NODE **mk_hashtab(void)
{
    return (NODE **) allocate(TAB_SIZE, sizeof(NODE *));
}

void set(NODE **hashtab, UINT x, UINT y)
{
    sets++;
    insert(&hashtab[HASH(x)], x, y);
}

UINT get(NODE **hashtab, UINT x)
{
    getss++;
    return find(hashtab[HASH(x)], x);
}

void print_hashtab(NODE **hashtab)
{
    int i;
    for (i=0; i<TAB_SIZE; i++)
    {
        printf("%u: ", i);
        print_list(hashtab[i]);
    }
}

main(int argc, char *argv[])
{
#ifdef TEST_LINKLIST
    int i;
    NODE *list = mk_list();
    for (i=0; i<10; i++)
        insert(&list, i, i);
    print_list(list);
    for (i=0; i<10; i++)
        printf("%u\n", find(list, i));
#else
    // default : test hash table 
    UINT long i;
    NODE **hashtab = mk_hashtab();
    UINT long num_node = (argc < 2) ? 21111 : atol(argv[1]);
    clock_t c1, c2;
    
    TAB_SIZE = (argc < 3) ? 1021 : atoi(argv[2]);
    
    srand(time(NULL));
    
    printf("Creating hash table with %lu random entries...", num_node);
    for (i=0; i<num_node; i++)
        set(hashtab, rand(), rand());
    printf("done\n");
    
    // print_hashtab(hashtab);

    printf("Doing %lu random searches...", num_node * 100);
    c1 = clock();
    for (i=0; i<num_node*100; i++)
        if (! get(hashtab, rand()))
            not_found++;
    c2 = clock();
    printf("done\n\n");
        
    printf("Table size:       %u\n",  TAB_SIZE);
    printf("List insertions:  %lu\n", inserts);
    printf("List finds:       %lu\n", finds);
    printf("List find/whiles: %lu\n", find_whiles);
    printf("Max whiles:       %u\n",  max_whiles);
    printf("Average whiles:   %lu\n", find_whiles / finds);
    printf("Hash sets:        %lu\n", sets);
    printf("Hash gets:        %lu\n", getss);
    printf("Not found:        %lu\n", not_found);
    printf("Clocks:           %lu\n", c2 - c1);
    printf("Seconds:          %f\n",  (double) (c2 - c1) / CLOCKS_PER_SEC);
    printf("Bytes allocated:  %lu\n", mem_alloc);
    printf("Items allocated:  %lu\n", mem_items);
    
    if (sets != inserts) puts("sets != inserts");
    if (getss != finds)  puts("getss != finds");
    if (inserts != num_node) puts("inserts != num_node");
    
    if (((find_whiles / finds) * TAB_SIZE) > finds)
        puts("Hash table not well distributed!!");
    
#endif /* TEST_LINKLIST */
}
