/*
MDFAT.C -- Dump a DoubleSpace MDFAT
Based on Microsoft's DSDUMP sample code
Andrew Schulman, July 1993
From Undocumented DOS, 2nd edition (Addison-Wesley, 1993)
NOTE:  *NONE* of this is undocumented!
*/

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <dos.h>

typedef unsigned char BYTE;
typedef unsigned short WORD;
typedef unsigned long DWORD;
typedef int BOOL;

typedef struct {
    BYTE    jmpBOOT[3];     // Jump to bootstrap routine
    char    achOEMName[8];  // OEM Name ("MSDSP6.0")
    // MS-DOS BPB
    WORD    cbPerSec;
    BYTE    csecPerClu;
    WORD    csecReserved;
    BYTE    cFATs;
    WORD    cRootDirEntries, csecTotalWORD;
    BYTE    bMedia;
    WORD    csecFAT, csecPerTrack, cHeads;
    DWORD   csecHidden, csecTotalDWORD;
    // DoubleSpace extensions
    WORD    secMDFATStart;
    BYTE    nLog2cbPerSec;
    WORD    csecMDReserved, secRootDirStart, secHeapStart, cluFirstData;
    BYTE    cpageBitFAT;
    WORD    RESERVED1;
    BYTE    nLog2csecPerClu;
    WORD    RESERVED2;
    DWORD   RESERVED3, RESERVED4;
    BYTE    f12BitFAT;
    WORD    cmbCVFMax;
    } MDBPB;

void fail(const char *s) { puts(s); exit(1); }

#define MDFAT_PER_BLOCK     (2048)
#define MDFAT_BLOCK_SIZE    (MDFAT_PER_BLOCK * sizeof(DWORD))

// handy macros from Microsoft's DSDUMP.C
#define GET_SECSTART(dw)    ((dw) & 0x3FFFFFL)
#define GET_CSEC_CODED(dw)  (1 + (int) (15 & ((dw) >> 22)))
#define GET_CSEC_PLAIN(dw)  (1 + (int) (15 & ((dw) >> 26)))
#define GET_FLAGS(dw)       (3 & ((dw) >> 30))
#define USED(dw)            (GET_FLAGS(dw) & 2)
#define UNCOMPRESSED(dw)    (GET_FLAGS(dw) & 1)

static DWORD sectors_coded = 0, sectors_plain = 0;
static WORD clusters_in_use = 0;
static WORD first_data_clust = 0;
static int verbose = 0, did_banner = 0;

void do_mdfat_entry(WORD cluster, DWORD mdfat_entry)
{
    int coded, plain;
    
    if (USED(mdfat_entry)) clusters_in_use++;
    else                   return;
    
    sectors_coded += (coded = GET_CSEC_CODED(mdfat_entry));
    sectors_plain += (plain = GET_CSEC_PLAIN(mdfat_entry));
    
    if (verbose || ( coded > plain))
    {
        if (! did_banner)
        {
            if (coded > plain)
                printf("Expanded clusters (cCmp > cUnc):\n");
            printf("Cluster    Sector cCmp cUnc C/U\n");
            printf("------     ------ ---- ---- ---\n");
            did_banner++;
        }
        printf("%5u   %8lu   %2u   %2u   %c\n", 
            cluster - first_data_clust,
            GET_SECSTART(mdfat_entry),
            coded, plain, 
            UNCOMPRESSED(mdfat_entry) ? 'U' : 'C');
    }
}

main(int argc, char *argv[])
{
    char *cvf_name;
    FILE *cvf;
    MDBPB mdbpb;
    DWORD *mdfat_block, num_sect, mdfat_bytes;
    WORD num_clust, mdfat_blocks, ratio, drive, i, j;
    
    if (argc < 2) fail("usage: mdfat [CVF file]");
    if (strcmp(strupr(argv[1]), "-VERBOSE") == 0)
        { verbose++; argv++; argc--; }
    cvf_name = argv[1];
        
    if ((cvf = fopen(cvf_name, "rb")) == NULL) fail("can't open CVF file");
    if (! fread(&mdbpb, sizeof(mdbpb), 1, cvf)) fail("can't read MDBPB");
    if ((mdbpb.jmpBOOT[0] != 0xE9) && (mdbpb.jmpBOOT[0] != 0xEB))
        fail("not a valid CVF file");   // has to start with JMP
    
    first_data_clust = mdbpb.cluFirstData;
    num_sect = (mdbpb.csecTotalWORD) ? 
        mdbpb.csecTotalWORD : mdbpb.csecTotalDWORD;
    num_clust = 1 + (num_sect / mdbpb.csecPerClu);
    mdfat_bytes = (DWORD) num_clust * sizeof(DWORD);
    mdfat_blocks = 1 + (mdfat_bytes / MDFAT_BLOCK_SIZE);
    
    if (! (mdfat_block = (DWORD *) malloc(MDFAT_BLOCK_SIZE)))
        fail("insufficient memory");
    
    fseek(cvf, (1 + mdbpb.secMDFATStart) * mdbpb.cbPerSec, SEEK_SET);

    for (i=0; i<mdfat_blocks-1; i++)
        if (fread(mdfat_block, MDFAT_PER_BLOCK, sizeof(DWORD), cvf))
            for (j=0; j<MDFAT_PER_BLOCK; j++)
                do_mdfat_entry((i*MDFAT_PER_BLOCK)+j, mdfat_block[j]);
        else
            fail("can't read MDFAT");
    
    if (fread(mdfat_block, MDFAT_PER_BLOCK, sizeof(DWORD), cvf))
        for (j=0; j<(num_clust % MDFAT_PER_BLOCK); j++)
            do_mdfat_entry((mdfat_blocks-1) + j, mdfat_block[j]);
    else
        fail("can't read MDFAT");   // okay to read less
        
    fclose(cvf);

    printf("\nTotal clusters:            %u\n", num_clust);
    printf("In use:                    %u (%lu%% used)\n", 
        clusters_in_use, (100L * clusters_in_use) / num_clust);
    printf("Plain sectors (cUnc):      %lu\n", sectors_plain);
    printf("Compressed sectors (cCmp): %lu (%lu%% compressed)\n", 
        sectors_coded, (100L * sectors_coded) / sectors_plain);
    ratio = (10L * sectors_plain) / sectors_coded;
    printf("Sector compression ratio:  %u.%u to 1\n",
        ratio / 10, ratio % 10);
    ratio = (160L * clusters_in_use) / sectors_coded;
    printf("DIR /C compression ratio:  %u.%u to 1\n",
        ratio / 10, ratio % 10);
    // to produce DIR /CH need get_dpb(host)->sectors_per_cluster
    
    return 0;
}
