/* 
MEMPROBE.C
Win32 Console application to test memory protection
Uses structured exception handling (SEH) to detect invalid memory access
Andrew Schulman (76320.302@compuserve.com), March 1995

Windows 95 doesn't just give each Win32 process full read/write access to
most of the shared lower 1MB of memory.  It gets worse!  It turns out
that Windows 95 also gives each Win32 process full read/write access
to many other parts of the system, such as the GDT and VMM service table!

I used Alex Shmidt's GATEVIEW to examine the Win95 page tables on my
system, and found 28 4MB regions marked as "System."  Of these, most
do appear to be inaccessible from a Win32 app.  However, at least the
following include globally-accessible read/write areas:

000:    0016B000    C133A000    00000000 - 003FFFFF  P  W   A   U   D   Sys 
300:    002E0000    C0F3C000    C0000000 - C03FFFFF  P  W   A   U   D   Sys 
305:    0016B000    C133A000    C1400000 - C17FFFFF  P  W   A   U   D  Sys VM1
30b:    00FC5000    C2F87000    C2C00000 - C2FFFFFF  P  W   A   U   D  Sys VM2
30c:    004A9000    C2FC8000    C3000000 - C33FFFFF  P  W   A   U   D  Sys VM3
30d:    00867000    C3201000    C3400000 - C37FFFFF  P  W   A   U   D  Sys VM4

The VM1,2,3,4 areas represent the "high linear" area for each VM that
happened to be running.  

VM1 is the System VM.  MEMPROBE is a Win32 Console app, so it runs in
VM1, but uses a different VM for its output.  In this test, MEMPROBE was
started from the DOS command line in VM2, and consequently was using VM2 for 
its output.  Here are MEMPROBE results in this configuration:

I then ran MEMPROBE on these areas:

;;; Lower 4MB of System VM
00001000h-00124000h Read/Write (1191936 bytes)
0019F000h-001A0000h Read/Write (4096 bytes)
73 seconds

;;; High linear of System VM
C1400000h-C1524000h Read/Write (1196032 bytes)
C159F000h-C15A0000h Read/Write (4096 bytes)
71 seconds

Notice that, while the bottom 4K are inaccessible at offset 0, 
presumably for NULL pointer detection, the same area of memory is
available via its high-linear address at C1400000h.

What is the 4K region at 19F000h?  Running another Win32 Console app
(DUMP) shows that it appears to include a thread control block
(THCB), among other things:

C159F050 | 01 00 00 00 01 00 00 00 00 00 00 00 A8 BE FE C0 | ................
C159F060 | 00 00 01 00 60 F0 D9 C2 B8 7E FE C0 54 48 43 42 | ....`....~..THCB
C159F070 | 70 EF 39 C1 E8 00 52 C1 01 00 2F 01 F6 11 00 00 | p.9...R.../.....
C159F080 | A7 75 00 00 00 00 00 00 17 01 AF 00 2A 00 00 00 | .u..........*...

;;; High linear of VM2 (used for this Win32 Console app's output)
C2C00000h-C2C02000h Read/Write (8192 bytes)
C2C02000h-C2C12000h Read
C2C12000h-C2CC0000h Read/Write (712704 bytes)
C2CC0000h-C2CC8000h Read
C2CC8000h-C2CE8000h Read/Write (131072 bytes)
C2CE8000h-C2CED000h Read
C2CED000h-C2CEE000h Read/Write (4096 bytes)
C2CEE000h-C2CFF000h Read
C2CFF000h-C2D24000h Read/Write (151552 bytes)
C2D9F000h-C2FE9000h Read/Write (2400256 bytes)
20 seconds

Presumably so much of this VM's address space is accessible because
this is the VM used for MEMPROBE's output?

;;; High linear of VM3 (an unrelated DOS box)
C3000000h-C3012000h Read
C3012000h-C30C0000h Read/Write (712704 bytes)
C30C0000h-C30C8000h Read
C30C8000h-C30E8000h Read/Write (131072 bytes)
C30E8000h-C30ED000h Read
C30ED000h-C30EE000h Read/Write (4096 bytes)
C30EE000h-C30FF000h Read
C30FF000h-C3124000h Read/Write (151552 bytes)
C319F000h-C3222000h Read/Write (536576 bytes)
63 seconds

;;; High linear of VM4 (another unrelated DOS box)
C3400000h-C3412000h Read
C3412000h-C34C0000h Read/Write (712704 bytes)
C34C0000h-C34C8000h Read
C34C8000h-C34E8000h Read/Write (131072 bytes)
C34E8000h-C34ED000h Read
C34ED000h-C34EE000h Read/Write (4096 bytes)
C34EE000h-C34FF000h Read
C34FF000h-C3524000h Read/Write (151552 bytes)
C359F000h-C35A0000h Read/Write (4096 bytes)
77 seconds

Why should these areas of totally unrelated DOS boxes be accessible to
a Win32 Console app?

I verified that all this is accessible from a non-Console Win32 app,
by compiling with -DNOT_CONSOLE.  Yes, the same results, except
memory access was significantly slower (presumably because all five
copies of MEMPROBE were running simultaneously), and except for this
one difference:

From non-Console app, couldn't access this chunk in VM2:
C2C00000h-C2C02000h Read/Write (8192 bytes)

It gets worse: It turns odd that there are other, even more vital,
areas to which a Win32 app has read/write access.  Running GATEVIEW
provides addresses for the page directory (which is specific to each
Win32 process context, presumably), for the GDT, the TSS, VMM's
service table, etc.  In this particular configuration, for example:

Page Directory    Physical = 002DE000   Linear = C0F3A000
GDTBase = C0034504   Limit = 01F7
TR = 0018   TSSBase = C000A78C  TSSLimit = 2069
VxD Root (VMM DDB pointer)  =   C000E248
VMM     0001    4.0 C00024B8    C0002B79    C0002B79    C000DC18    0000018C

Turns out these are *ALL* read/write accessible to a Win32 process!

C0001000h-C0035000h Read/Write (212992 bytes)
C0200000h-C0203000h Read/Write (12288 bytes)
C0220000h-C0288000h Read/Write (425984 bytes)
C0320000h-C0327000h Read/Write (28672 bytes)
C03E0000h-C03E2000h Read/Write (8192 bytes)
C0ECE000h-C0EDA000h Read/Write (49152 bytes)
C0EEA000h-C0EF6000h Read/Write (49152 bytes)
C0F1A000h-C0F9C000h Read/Write (532480 bytes)
C0FCE000h-C0FF3000h Read/Write (151552 bytes)
C10CD000h-C116F000h Read/Write (663552 bytes)
C11CD000h-C135F000h Read/Write (1646592 bytes)
C1380000h-C1381000h Read/Write (4096 bytes)
C139C000h-C13A1000h Read/Write (20480 bytes)
C13AF000h-C13B2000h Read/Write (12288 bytes)
C13B3000h-C13B4000h Read/Write (4096 bytes)
C13B8000h-C13B9000h Read/Write (4096 bytes)
C13C1000h-C13C2000h Read/Write (4096 bytes)
C13C9000h-C13D4000h Read/Write (45056 bytes)
C13F3000h-C13FE000h Read/Write (45056 bytes)
....

For example, GATEVIEW shows VMM service table at C000DC18h.  Can view
this from a Win32 app, DUMP:

C000DC18 | 70 2B 00 C0 D3 1F 00 C0 DA 1F 00 C0 E1 1F 00 C0 | p+..............
C000DC28 | E8 1F 00 C0 F0 1F 00 C0 F8 42 00 C0 FF 42 00 C0 | .........B...B..
C000DC38 | 08 43 00 C0 E8 37 22 C0 70 38 22 C0 24 1A 00 C0 | .C...7".p8".$...
C000DC48 | 76 1A 00 C0 28 1B 00 C0 EB 46 01 C0 0C 2C 00 C0 | v...(....F...,..
...

Can confirm by running VXDLIST -VERBOSE:

   010000h @ C0002B70h   Get_VMM_Version
   010001h @ C0001FD3h   Get_Cur_VM_Handle
   010002h @ C0001FDAh   Test_Cur_VM_Handle
   010003h @ C0001FE1h   Get_Sys_VM_Handle
   010004h @ C0001FE8h   Test_Sys_VM_Handle

Can we actually write to these?  Yes, confirmed with RW:

Try to read/write 273E8h (160744) bytes from C000DC18h to C0035000h
Successfully read/wrote 273E8h (160744) bytes from C000DC18h to C0035000h

Was taking forever to run MEMPROBE, so had increment by pages (1024) not
ints (1).  MEMPROBE.LOG:

00001000h-00124000h Read/Write (1191936 bytes)
0019F000h-001A0000h Read/Write (4096 bytes)
00400000h-00406000h Read
00406000h-00408000h Read/Write (8192 bytes)
00408000h-00409000h Read
00409000h-0040B000h Read/Write (8192 bytes)
0040B000h-0040C000h Read
00410000h-00411000h Read/Write (4096 bytes)
00510000h-00511000h Read/Write (4096 bytes)
00520000h-00521000h Read/Write (4096 bytes)
00637000h-00638000h Read/Write (4096 bytes)
0063E000h-00640000h Read/Write (8192 bytes)
00650000h-00654000h Read/Write (16384 bytes)
00660000h-00661000h Read/Write (4096 bytes)
80001000h-80091000h Read/Write (589824 bytes)
8009E000h-8015D000h Read/Write (782336 bytes)
8015E000h-80168000h Read/Write (40960 bytes)
80169000h-8016A000h Read/Write (4096 bytes)
8016B000h-8017F000h Read/Write (81920 bytes)
80180000h-8018B000h Read/Write (45056 bytes)
801F8000h-8028C000h Read/Write (606208 bytes)
81523000h-81525000h Read/Write (8192 bytes)
81535000h-81542000h Read/Write (53248 bytes)
81634000h-81636000h Read/Write (8192 bytes)
81A34000h-81A36000h Read/Write (8192 bytes)
81A39000h-81A3A000h Read/Write (4096 bytes)
81A3C000h-81A3D000h Read/Write (4096 bytes)
81A3D000h-81A67000h Read
81A67000h-81A79000h Read/Write (73728 bytes)
81A88000h-81A8B000h Read/Write (12288 bytes)
81C87000h-81C8D000h Read/Write (24576 bytes)
81D07000h-81D19000h Read/Write (73728 bytes)
81D28000h-81D2A000h Read/Write (8192 bytes)
81F27000h-81F29000h Read/Write (8192 bytes)
81F38000h-81F3D000h Read/Write (20480 bytes)
82137000h-82139000h Read/Write (8192 bytes)
82139000h-8213E000h Read
8213E000h-82155000h Read/Write (94208 bytes)
82156000h-82159000h Read/Write (12288 bytes)
8215C000h-8215D000h Read/Write (4096 bytes)
82166000h-82176000h Read/Write (65536 bytes)
82374000h-82376000h Read/Write (8192 bytes)
82475000h-8247D000h Read/Write (32768 bytes)
82675000h-826FE000h Read/Write (561152 bytes)
82700000h-82788000h Read/Write (557056 bytes)
82790000h-827A2000h Read
82BBF000h-833BF000h Read/Write (8388608 bytes)
BFEE0000h-BFEE2000h Read
BFEE2000h-BFEE3000h Read/Write (4096 bytes)
BFEE3000h-BFEE8000h Read
BFEF0000h-BFF14000h Read
BFF14000h-BFF15000h Read/Write (4096 bytes)
BFF15000h-BFF22000h Read
BFF30000h-BFF4A000h Read
BFF4A000h-BFF4C000h Read/Write (8192 bytes)
BFF4C000h-BFF50000h Read
BFF50000h-BFF51000h Read/Write (4096 bytes)
BFF51000h-BFF54000h Read
BFF60000h-BFF67000h Read
BFF67000h-BFF68000h Read/Write (4096 bytes)
BFF68000h-BFF6F000h Read
BFF70000h-BFFB7000h Read
BFFB9000h-BFFBD000h Read
BFFBD000h-BFFC1000h Read/Write (16384 bytes)
BFFC1000h-BFFC6000h Read
BFFC6000h-BFFC9000h Read/Write (12288 bytes)
BFFC9000h-BFFDB000h Read
C0001000h-C0091000h Read/Write (589824 bytes)
C0200000h-C0203000h Read/Write (12288 bytes)
C0220000h-C0288000h Read/Write (425984 bytes)
C0320000h-C0327000h Read/Write (28672 bytes)
C03E0000h-C03E2000h Read/Write (8192 bytes)
C0ECE000h-C0ECF000h Read/Write (4096 bytes)
C0ED0000h-C0ED1000h Read/Write (4096 bytes)
C0ED2000h-C0ED3000h Read/Write (4096 bytes)
C0ED4000h-C0ED5000h Read/Write (4096 bytes)
C0ED6000h-C0ED7000h Read/Write (4096 bytes)
C0ED8000h-C0ED9000h Read/Write (4096 bytes)
C0EDA000h-C0EDB000h Read/Write (4096 bytes)
C0EDC000h-C0EDD000h Read/Write (4096 bytes)
C0EDE000h-C0EDF000h Read/Write (4096 bytes)
C0EE0000h-C0EE1000h Read/Write (4096 bytes)
C0EE2000h-C0EE3000h Read/Write (4096 bytes)
C0EE4000h-C0EE5000h Read/Write (4096 bytes)
C0EF6000h-C0EF7000h Read/Write (4096 bytes)
C0EF8000h-C0EF9000h Read/Write (4096 bytes)
C0EFA000h-C0EFC000h Read/Write (8192 bytes)
C0EFD000h-C0EFE000h Read/Write (4096 bytes)
C0FCE000h-C0FEF000h Read/Write (135168 bytes)
C10CD000h-C1166000h Read/Write (626688 bytes)
C11CD000h-C135F000h Read/Write (1646592 bytes)
C1380000h-C1381000h Read/Write (4096 bytes)
C139C000h-C23DE000h Read/Write (17047552 bytes)
C23DF000h-C23E2000h Read/Write (12288 bytes)
C23F0000h-C23F3000h Read/Write (12288 bytes)
C23F4000h-C23F5000h Read/Write (4096 bytes)
C23F9000h-C23FA000h Read/Write (4096 bytes)
C23FD000h-C2524000h Read/Write (1208320 bytes)
C259F000h-C25A0000h Read/Write (4096 bytes)
C3190000h-C319B000h Read/Write (45056 bytes)
C31BA000h-C31BB000h Read/Write (4096 bytes)
C31CA000h-C31D5000h Read/Write (45056 bytes)
C3400000h-C3401000h Read/Write (4096 bytes)
C3401000h-C340F000h Read
C340F000h-C3410000h Read/Write (4096 bytes)
C3410000h-C3416000h Read
C3416000h-C34C0000h Read/Write (696320 bytes)
C34C0000h-C34C8000h Read
C34C8000h-C34E8000h Read/Write (131072 bytes)
C34E8000h-C34ED000h Read
C34ED000h-C34EE000h Read/Write (4096 bytes)
C34EE000h-C34FF000h Read
C34FF000h-C3524000h Read/Write (151552 bytes)
C359F000h-C35A0000h Read/Write (4096 bytes)
C3800000h-C3816000h Read
C3816000h-C38C0000h Read/Write (696320 bytes)
C38C0000h-C38C8000h Read
C38C8000h-C38E8000h Read/Write (131072 bytes)
C38E8000h-C38ED000h Read
C38ED000h-C38EE000h Read/Write (4096 bytes)
C38EE000h-C38FF000h Read
C38FF000h-C3924000h Read/Write (151552 bytes)
C399F000h-C39A0000h Read/Write (4096 bytes)
409 seconds

Number and percent of pokeable bytes depends largely on how many VMs
running, and of course on size of app.  But generally around 1 percent
of address space is pokeable by MEMPROBE.
*/


#include <stdlib.h>
#include <stdio.h>
#include <time.h>
#define WIN32_LEAN_AND_MEAN
#include "windows.h"

#define USE_SEH

typedef enum { 
    NONE = 0, BAD_READ=1, BAD_WRITE=2, BAD_RW=3, GOOD_RW=4,
    } rgn;

FILE *logfile;

#ifdef NOT_CONSOLE
#define PUTS(s)         MessageBox(0, (s), "MEMPROBE", MB_OK)
#else
#define PUTS(s)         puts(s)
#endif

void fail(const char *s) { PUTS(s); exit(1); }

void display_range(rgn state, int *start, unsigned size)
{
    static char *name[5] = { 
        "Internal error!", "Write", "Read", "Protected", "Read/Write"
        } ;
    unsigned ul = (unsigned) start;
    
    if (! state) return;
    if (state == BAD_RW) return;    // don't show "Protected"
        
    fprintf(logfile, "%08lXh-%08lXh\t%s", 
        ul, ul+(size*sizeof(int)), name[state]);
    if (state == GOOD_RW) fprintf(logfile," (%lu bytes)", size*sizeof(int));
    fprintf(logfile,"\n");
    fflush(logfile);
#ifndef NOT_CONSOLE 
    fprintf(stderr, "%08lXh-%08lXh\t%s", 
        ul, ul+(size*sizeof(int)), name[state]);
    if (state == GOOD_RW) fprintf(stderr," (%lu bytes)", size*sizeof(int));
    fprintf(stderr,"\n");
#endif
}

#ifdef NOT_CONSOLE
#define argc __argc
#define argv __argv
extern int __argc;
extern char **__argv;

int PASCAL WinMain(HANDLE hInstance, HANDLE hPrevInstance,
    LPSTR lpszCmdLine, int nCmdShow)
#else
main(int argc, char *argv[])
#endif
{
    unsigned size, pokeable_bytes = 0;
    rgn state = NONE, newst = NONE;
    int r, w, x, *p, *start, *end, *begin;
    char *filename;
    time_t t1, t2;
    
    if (argc < 2) p = (int *) 0;
    else sscanf(argv[1], "%x", (unsigned *) &p);  
    begin = p;
    if (argc < 3) end = (int *) 0;
    else sscanf(argv[2], "%x", (unsigned *) &end);
    filename = (argc < 4) ? "memprobe.log" : argv[3];
    logfile = fopen(filename, "w");
    
    if (! logfile) fail("Can't create log file");
    
    time(&t1);
    
    do {
#ifdef USE_SEH
        // use Structured Exception Handling (SEH)
        _try { r = 0; x = *p; }                             // try peek
        _except (EXCEPTION_ACCESS_VIOLATION) { r = 1; }     // peek failed
            
        // because we only write to write if we could read,
        // the BAD_READ state never happens

        if (r == 0)     // if could successfully peek
        {
            _try { w = 0; x = *p; *p = x; }                 // try poke
            _except (EXCEPTION_ACCESS_VIOLATION) { w = 1; } // poke failed
            if (*p != x) w = 1;                    // didn't really write?
        }
        else w = 1;     // if not readable, assume not writeable
#else
        // use the Win32 IsBadXXXXPtr API functions
        if ((r = IsBadReadPtr(p, sizeof(int))) == 0)    // if readable
            w = IsBadWritePtr(p, sizeof(int));          // see if writeable
        else w = 1;
#endif      
        
        newst = (r && w) ? BAD_RW : r ? BAD_READ : w ? BAD_WRITE : GOOD_RW;
        
        if (newst == state)
            size += 1024;
        else
        {
            display_range(state, start, size);
            
            if (state == GOOD_RW) pokeable_bytes += (size * sizeof(int));
            
            start = p;
            size = 1024;
            state = newst;
        }

        p += 1024;      // do by pages?
    } while (p != end);
    
    time(&t2);
    
    // get last one
    display_range(state, start, size);
    
    fprintf(logfile, "%lu seconds\n", t2-t1);
    fprintf(logfile, "%lu pokeable bytes out of %lu\n", 
        pokeable_bytes,
        (end-begin) * sizeof(int));
    fprintf(logfile, "%0.2f%% pokeable\n",
        (pokeable_bytes * 100.0) / ((end-begin) * sizeof(int)));
#ifdef NOT_CONSOLE
    MessageBox(0, "Done!", "MEMPROBE", MB_OK);
#else
    fprintf(stderr, "%lu seconds\n", t2-t1);
#endif
    
    fclose(logfile);

    return 0;
}
