/*
MSDETECT.C
Andrew Schulman, May 1993
From "Undocumented DOS", 2nd edition (Addison-Wesley, 1993)
See also Dr. Dobb's Journal, September 1993

A duplication of Microsoft's MS-DOS detection code from Windows 3.1
WIN.COM, SMARTDRV.EXE, HIMEM.SYS, SETUP.EXE.

The original Microsoft code (with the initials 'AARD') is heavily XOR
encrypted and obfuscated.  Here the encryptions and obfuscations have
been removed.  The AARD code also attempts to disable debuggers by
pointing INT 1, 2, and 3 at invalid code (FFh FFh); this also has
been avoided here.  

The original AARD code produces the following message:
    "Non-Fatal error detected: error #2726
    Please contact Windows 3.1 beta support
    Press ENTER to exit or C to continue"
This has been replaced with more informative messages about the
exact "problem" (i.e., difference from MS-DOS).

Geoff Chappell (geoffc@cix.compulink.co.uk) deciphered the original
code's tests (upper case map segment, FCB-SFT) in the case where the
preliminary SysVars tests fail.  Some of this material is discussed
in Geoff's forthcoming book, "DOS Internals" (Addison-Wesley, 1993).

Optional command-line switches -DOREDIR and -NOREDIR have been added
to aid testing both crucial tests whether or not a redirector is
actually installed.

Build program with Microsoft C: cl msdetect.c
*/

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <dos.h>

typedef int BOOL;
typedef unsigned char BYTE;
typedef unsigned short WORD;
typedef unsigned long DWORD;
typedef void far *FP;

BYTE far *_dos_getsysvars(void);
FP _dos_getcasemap(void);
WORD _dos_getdataseg(void);
BOOL _dos_isredirector(void);

void fail(const char *s) { puts(s); exit(1); }

main(int argc, char *argv[])
{
    BYTE far *sysvars;
    int do_redir = -1;
    
    // this is just to make testing easier
    // can fake presence or absence of a network redirector
    // to exercise both tests
    if (argc > 1 && argv[1][0] == '-')
    {
        char *s = strupr(&argv[1][1]);
        if (strcmp(s, "DOREDIR") == 0)
            do_redir = 1;
        else if (strcmp(s, "NOREDIR") == 0)
            do_redir = 0;
        else
            fail("usage: msdetect [-doredir] [-noredir]");
    }

    if ((sysvars = _dos_getsysvars()) == 0)
        fail("INT 21h AX=5200h returns 0!");
    
    if (_osmajor >= 0x0a)
        fail("DOS version >= 10; this is OS/2 (or early NT beta!)");
    
    #define SYSVARS(ofs)            (*((FP far *) &sysvars[ofs]))
    #define SYSVARS_TEST(ofs, msg)  if (! SYSVARS(ofs)) fail(msg)
        
    SYSVARS_TEST(0, "Disk Parameter Block (DPB) pointer in SysVars is 0!");
    SYSVARS_TEST(4, "System File Table (SFT) pointer in SysVars is 0!");
    SYSVARS_TEST(8, "CLOCK$ device pointer in SysVars is 0!");
    SYSVARS_TEST(0x12, "buffers header pointer in SysVars is 0!");
    
    // this test fails in device-driver init (HIMEM.SYS) under DR DOS 6.0
    SYSVARS_TEST(0x16, 
        "Current Directory Structure (CDS) pointer in SysVars is 0!");

    SYSVARS_TEST(0x0C, "CON device pointer in SysVars is 0!");
    SYSVARS_TEST(0x22, "Device chain pointer (from NUL) in SysVars is 0!");
    
    if ((! _dos_isredirector()) && (do_redir == -1))
        do_redir = 0;

    // the following tests fail under DR DOS 5 and 6
    // (and under beta of Novell DOS 7)
    if (do_redir)
    {
        FP casemap;
        puts("Doing redirector test");
        casemap = _dos_getcasemap();
        if (FP_SEG(casemap) != _dos_getdataseg())
            fail("Default case map isn't in DOS data segment!");
        printf("case map @ %Fp\n", casemap);
    }
    else
    {
        puts("Doing no-redirector test");
        if (FP_OFF(SYSVARS(0x1A)) != 0)
            fail("First FCB-SFT not located on paragraph boundary!");
        printf("FCB-SFT ptr @ %Fp -> %Fp\n", sysvars+0x1a, SYSVARS(0x1A));
    }
    
    // if get here, everything checks out
    puts("All tests check out: must be MS-DOS");
    return 0;
}

// undocumented function
BYTE far *_dos_getsysvars(void)
{
    // could initialize ES:BX to 0:0 but the MS code doesn't do this
    _asm mov ax, 5200h
    _asm int 21h
    _asm mov dx, es
    _asm mov ax, bx
    // ES:BX retval moved into DX:AX
}

// see "Microsoft MS-DOS Programmer's Reference", p. 143
// formerly undocumented
BOOL _dos_isredirector(void)
{
    BYTE retval;
    _asm mov ax, 1100h
    _asm int 2fh
    _asm mov retval, al
    return (retval == 0xFF);
}

// undocumented function
WORD _dos_getdataseg(void)
{
    _asm push ds
    _asm mov ax, 1203h
    _asm int 2fh
    _asm mov ax, ds
    _asm pop ds
    // retval in AX
}

// get a far pointer to the default case map
// see "Microsoft MS-DOS Programmer's Reference", pp. 272-3
FP _dos_getcasemap(void)
{
    BYTE country_info[34];
    FP fp = (FP) country_info;
    _asm push ds
    _asm mov ax, 3800h
    _asm lds dx, dword ptr fp
    _asm int 21h
    _asm pop ds
    return *((FP far *) &country_info[18]);
}
