/* 
NAMCLUST.C -- Convert file name to starting cluster
Andrew Schulman, July 1993
From Undocumented DOS, 2nd edition (Addison-Wesley, 1993)

Overall structure:
main
	truename -- to get canonical filename
	get_dpb
	do_cluster -- process a directory cluster; start with root dir
		read_cluster	
			read_sectors -- use INT 25h to read
		do_dir
			printf -- success!  print results
			do_cluster -- partial match; recurse to next level
			get_fat_entry -- get next cluster of directory
		fail -- no such file!
*/

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <ctype.h>
#include <dos.h>
#include "diskstuf.h"

// Hummel, Assembly Language Lab Notes, p. 88
#define CLUSTER_TO_LSN(clust)   \
    ((((DWORD) (clust) - 2) * sect_per_clust) + first_sect)

void do_cluster(WORD cluster);
void do_dir(WORD cluster, DIR_ENTRY *buffer);
void read_cluster(int drive, BYTE far *buffer, WORD cluster);

static STRING glob_name[64] = {0};
static DPB far *glob_dpb;
static WORD want_cluster = 0xFFFF;
static WORD bytes_per_clust = 0, sect_per_clust = 0, first_sect = 0;
static int level = 0, glob_num_dir = 0, glob_drive;
static int num_levels = 0;
static char canon[128];
static int do_all = 0;

void fail(const char *s) { puts(s); exit(1); }

main(int argc, char *argv[])
{
    char partname[64];
    char *s, *s2;
    int i, part;
	
	if (argc < 2)
		fail("usage: namclust [pathname]");
    
    for (i=0; i<64; i++)
        glob_name[i] = (STRING) malloc(16);
    
    // program uses DOS 3.31 style of INT 25h
    if ((_osmajor < 3) || (_osmajor == 3 && _osminor < 31))
        fail("This program requires DOS 3.31 or higher");
	
	if (strlen(argv[1]) == 2 && argv[1][1] == ':')
	{
		do_all++;
		num_levels = 0xFFFF;
		strcpy (canon, strupr(argv[1]));
		goto start;
	}
    
    if (! truename(argv[1], canon))
		fail("can't get truename");
    if (canon[1] != ':')
        fail("Sorry, this doesn't work on network-redirector drives");
    
    for (part=0, s=canon, s2=partname; ; s++)
        if ((*s == '\0') || (*s == '\\'))
        {
            *s2 = '\0';
            strcpy(glob_name[part], partname);
            part++;
            if (*s == '\0')
                break;
            s2 = partname;
        }
        else
            *s2++ = *s;

	num_levels = part - 1;
    level = 1;

start:
    glob_drive = canon[0] - 'A';
    if (! (glob_dpb = get_dpb(glob_drive+1)))
        fail("can't get DPB");
	
    sect_per_clust = glob_dpb->sectors_per_cluster + 1;
    bytes_per_clust = glob_dpb->bytes_per_sect * sect_per_clust;
    first_sect = glob_dpb->first_data_sector;
    glob_num_dir = bytes_per_clust / sizeof(DIR_ENTRY);  
	do_cluster(0);				// kick-start search
}

void do_cluster(WORD cluster)
{
    BYTE *buffer;
	if (cluster == 0)	// phony cluster # for root directory
	{
		int num_sect;
		buffer = (BYTE *) calloc(glob_dpb->max_root_dir, sizeof(DIR_ENTRY));
		if (! buffer)
			fail("insufficient memory");
		num_sect = (glob_dpb->max_root_dir * sizeof(DIR_ENTRY)) /
			glob_dpb->bytes_per_sect;
		read_sectors(glob_drive, buffer, num_sect, 
			(_osmajor >= 4) ? glob_dpb->vers.dos4.first_dir_sector :
						      glob_dpb->vers.dos3.first_dir_sector);
	}
	else
	{
		printf("(Read cluster %u)\n", cluster);
		if (! (buffer = (BYTE *) calloc(bytes_per_clust, 1)))
			fail("insufficient memory");
		read_cluster(glob_drive, buffer, cluster);
	}
    do_dir(cluster, (DIR_ENTRY *) buffer);
	if (do_all)
		free(buffer);
	else
		fail("can't find file");    // error if get here
}

void do_dir(WORD cluster, DIR_ENTRY *buffer)
{
    DIR_ENTRY *dir, *p;
    char name[16], *s;
    WORD next, num_dir;
    int entry, i;
	
    dir = (DIR_ENTRY *) buffer;
	num_dir = (cluster == 0) ? glob_dpb->max_root_dir : glob_num_dir;

    for (entry=0, p=dir; entry<num_dir; p++, entry++)
    {
        if (p->name[0] == '\0') return;     // end of dir
        if (p->name[0] == 0xE5) continue;   // deleted entry
        if (p->name[0] == '.')  continue;   // don't bother with . and ..
            
        memcpy(name, p->name, 8); name[8] = 0;
        s = name; while (*s && *s != ' ') s++; *s = '\0';
        if (name[0] == 0x05) name[0] = 0xE5;    // 0xE5 is valid first char
        if (p->ext[0] != ' ')
        {
            strcat(name, ".");
            strncat(name, p->ext, 3);   
            s = name; while (*s && *s != ' ') s++; *s = '\0';
        }

        if (do_all || strcmp(name, glob_name[level]) == 0)    // got match!
        {
            if (do_all || level == num_levels)				// done
            {
				DWORD sector = CLUSTER_TO_LSN(p->cluster);
				char *s;
				if (do_all)	// indent to show level
				{
					int i;
					for (i=0; i<level; i++)
						printf("   ");
					s = name;
				}
				else 
					s = canon;
				printf("%s ==> %u (sect %lu, 0x%08lx)\n", 
					s, p->cluster, sector, sector);
				if (! do_all)
					exit(0);
            }
            if (p->attrib & DIRECTORY_ATTR) // already know not . or ..
            {
                level++;
                do_cluster(p->cluster);     // if directory, recurse
				level--;					// for do_all
            }
        }
    }
	
    // locate next directory cluster
    if ((next = get_fat_entry(glob_drive, glob_dpb, cluster)) < 0xFFF0)
        do_cluster(next);
}

void read_cluster(int drive, BYTE far *buffer, WORD cluster)
{
    DWORD sector;
    if (cluster)            sector = CLUSTER_TO_LSN(cluster);
    else if (_osmajor >= 4) sector = glob_dpb->vers.dos4.first_dir_sector;
    else                    sector = glob_dpb->vers.dos3.first_dir_sector;
    if (! read_sectors(drive, buffer, sect_per_clust, sector))
        fail("can't read cluster");
}
        
