/* 
NODOSMGR.C
From "Undocumented DOS", 2nd edition (Addison-Wesley, 1993)
Andrew Schulman, June 1993
bcc -P- -2 nodosmgr.c

By preventing MS-DOS from seeing Windows calls to INT 2Fh AX=1607h
BX=15h, this program simulates the effect of running Windows on a
version of DOS 5+ which fails to implement the undocumented DOSMGR
interface.

Results from running on MS-DOS 6.0:
C:\UNDOC2\CHAP1>nodosmgr win
;;; Windows starts to come up
ERROR: Unsupported MS-DOS version.
5 calls to DOSMGR 2F/1607/15 interface

Added option to pass through specified functions:
nodosmgr win 0      -- ERROR: Unsupported               4 calls
nodosmgr win 1      -- Windows runs!                    6 calls
nodosmgr win 2      -- ERROR: Unsupported               5 calls
nodosmgr win 3      -- ERROR: Unsupported               5 calls
nodosmgr win 4      -- ERROR: Unsupported               5 calls
nodosmgr win 5      -- ERROR: Unsupported               5 calls

So, simply by implementing function 1 (2f/1607/15/1), can run
Windows.  This is "Enable Specialised Windows Support" (see Geoff
Chappell, "DOS Internals" mss., p. 61).  How much does DOSMGR care
about?  Just wants the weird A2AB:B97C signature in DX:AX?

Should be easy to implement func 1 without passing through to DOS.
Based on DEBUG disassembly of DOS 6.0 (\undoc2\chap6\int212f.out),
here is DOS handler for 2f/1607/15/1:

                        loc_4678:
FDC8:4678   8BDA            MOV BX,DX 
FDC8:467A   EB0F            JMP loc_468B 
; ...
                        loc_468B:
FDC8:468B   B87CB9          MOV AX,B97C 
FDC8:468E   BAABA2          MOV DX,A2AB 

Should be easy to emulate this.  Will Windows still run?  If so, then
DOSMGR interface appears to be somewhat artificial, the actual
"functionality" consisting mostly of magic numbers B97Ch and A2AB2!

Yes, Windows still runs!  Thus, the real "functionality" of DOSMGR
seems to reside in the magic numbers.  Perhaps Windows is only
running by "dumb luck," but it sure seems to work.  I tested running
several DOS boxes in the background, several Win apps, all at the
same time.  A DOS emulator can PERHAPS thus be DOS 5+ and run Windows
Enhanced mode simply by hooking 2F/1607/15/1 and returning A2AB/B97C
with BX=DX. 

This raises the possibility that real purpose of DOSMGR is to tie
Windows and DOS 5+ in an artificial way, and that the rest of the
interface is "window dressing," as it were.  This is simply a
possibility, but one that is worth examining.

This possibility is underlined by the fact that some pieces of the
interface are not implemented by MS-DOS, and that some implementations
are trivial:

MS-DOS 6.0 implementation:

func 0: ret cx=1, es:bx -> ds:1022 (DOS var patch table)
func 1: ret bx=dx, dx:ax=magic
func 2: not handled
func 3: if (dx & 1) return cx=58h (CDS entry size), dx:ax=magic
func 4: ret dx = 0 (trivial)
func 5: get device driver size from MCB; ret dx:ax=magic

Functions 0 and 5 have some substance, but appear to not be necessary?
(At least when running on DOS 6!!).  Why this stuff required of underlying
DOS when tons of much worse stuff still hard-coded into DOSMGR?  Because
had idea of decoupling DOSMGR from knowledge of DOS internals, or because
interface not necessary except to make Windows inhospitable to other
DOSes???

If DOSMGR interface not enabled, then DOS machine ID (VM #) at 0116:033E
is zero.  Does this really make any difference to DOS functioning?  
From int212f.lst, appears to be used e.g. related to SHARE?  Run SHARE
are try experiment again!  Didn't seem to make any difference.  Though
did have problems when closing down DOS boxes; repeatable?  
*/

#include <stdlib.h>
#include <stdio.h>
#include <process.h>
#include <dos.h>

#pragma pack(1)

typedef struct {
#ifdef __TURBOC__
    unsigned short bp,di,si,ds,es,dx,cx,bx,ax;
#else
    unsigned short es,ds,di,si,bp,sp,bx,dx,cx,ax;     /* same as PUSHA */
#endif
    unsigned short ip,cs,flags;
    } REG_PARAMS;
    
volatile unsigned dosmgr_calls = 0;
int pass_through[10] = { 0 };   /* list of DOSMGR calls ok to pass to DOS */

#define DOSMGR_MAGIC_DX     0xA2AB
#define DOSMGR_MAGIC_AX     0xB97C

void interrupt far int2f(REG_PARAMS r);
void (interrupt far *old)(void);

void fail(char *s) { puts(s); exit(1); }

main(int argc, char *argv[])
{
    int i, func;
    
    if (argc < 2) fail("usage: nodosmgr [win] <func list to pass through>");
    
    for (i=2; i<argc; i++)
        if ((func = atoi(argv[i])) < 10)
        {
            pass_through[func]++;
            argv[i] = (void *) 0;
        }
    
    old = (void (interrupt far *)(void)) _dos_getvect(0x2F); // hook 2F
    _dos_setvect(0x2F, int2f);

    spawnvp(P_WAIT, argv[1], &argv[1]);     // run command (e.g., Windows)
    
    _dos_setvect(0x2F, old);                // unhook 2F
    
    printf("%u calls to DOSMGR 2F/1607/15 interface\n", dosmgr_calls);
    return 0;
}

void interrupt far int2f(REG_PARAMS r)
{
    if ((r.ax == 0x1607) && (r.bx == 0x15))
    {
        dosmgr_calls++;
        if (pass_through[r.cx])
        {
            if (r.cx == 1)
            {
                r.bx = r.dx;                // do just what DOS does
                r.dx = DOSMGR_MAGIC_DX;     // but don't pass down to DOS
                r.ax = DOSMGR_MAGIC_AX;
            }
            else
                _chain_intr(old);
        }
        // otherwise, don't pass down to DOS
    }
    else
        _chain_intr(old);
}

