/*
PROT.C
Routines for protected mode programming with Windows and DPMI
Andrew Schulman, February 1993
from "Undocumented DOS", 2nd edition (Addison-Wesley, 1993)
(Changed version of PROTMODE, from MSJ, Oct. 1992, Dec. 1992)

Requires 286+ instruction set (e.g., bcc -2 or cl -G2)
*/

#include <stdlib.h>
#include <string.h>
#include <dos.h>
#ifndef DPMI_APP
#include "windows.h"
#endif
#include "prot.h"

/**********************************************************************/

static WORD mapped = 0;     // to ensure don't have selector leak

WORD get_mapped(void)   { return mapped; }

/**********************************************************************/

static DWORD base = 0;      // for data in other VMs

void set_base(DWORD b)  { base = b; }
DWORD get_base(void)    { return base; }

/**********************************************************************/

void far *map_linear(DWORD lin_addr, DWORD num_bytes)
{
    WORD sel;
    
    /* allocate a selector similar to our current DS 
       (i.e., a data selector) */
    _asm mov sel, ds
    if ((sel = AllocSelector(sel)) == 0)
        return (void far *) 0;
    
    /* set the base and limit of the new selector; variable "base"
       allows access to data in other VMs */
    SetSelectorBase(sel, base + lin_addr);
    SetSelectorLimit(sel, num_bytes - 1);   
    
    mapped++;
    
    /* turn into a far pointer */
    return MK_FP(sel, 0);
}

void free_mapped_linear(void far *fp)
{
    FreeSelector(FP_SEG(fp));
    mapped--;
}

void far *map_real(void far *fp, DWORD size)
{
    return map_linear(MK_LIN(fp), size);
}

/**********************************************************************/

/* Performs a real mode interrupt from protected mode */
#pragma warn -rvl
#pragma argsused
BOOL dpmi_rmode_intr(WORD intno, WORD flags, 
    WORD copywords, RMODE_CALL far *rmode_call)
{
    // ignore flags
    _asm  push di
    _asm  push bx
    _asm  push cx
    _asm  mov ax, 0300h       // simulate real mode interrupt
    _asm  mov bx, intno       // interrupt number, flags
    _asm  mov cx, copywords   // words to copy from pmode to rmode stack
    _asm  les di, rmode_call  // ES:DI = address of rmode call struct
    _asm  int 31h             // call DPMI
    _asm  jc error
    _asm  mov ax, 1           // return TRUE
    _asm  jmp short done
error:
    _asm mov ax, 0           // return FALSE
done:   
    _asm pop cx
    _asm pop bx
    _asm pop di
}

/**********************************************************************/

int real_int86x(int intno, union REGS *inregs, union REGS *outregs,
    struct SREGS *sregs)
{
    RMODE_CALL r;
    memset(&r, 0, sizeof(r)); // initialize all fields to zero: important!
    r.edi = inregs->x.di;   r.esi = inregs->x.si;
    r.ebx = inregs->x.bx;   r.edx = inregs->x.dx;
    r.ecx = inregs->x.cx;   r.eax = inregs->x.ax;
    r.flags = inregs->x.cflag;
    r.es = sregs->es;       r.ds = sregs->ds;
    r.cs = sregs->cs;       
    // NOTE: r.ss=r.sp=0 so that DPMI host provides real mode stack!
    if (! dpmi_rmode_intr(intno, 0, 0, &r))
    {
        outregs->x.cflag = 1;   // error: set carry flag!
        return 0;
    }
    sregs->es = r.es;       sregs->cs = r.cs;
    sregs->ss = r.ss;       sregs->ds = r.ds;
    outregs->x.ax = r.eax;  outregs->x.bx = r.ebx;
    outregs->x.cx = r.ecx;  outregs->x.dx = r.edx;
    outregs->x.si = r.esi;  outregs->x.di = r.edi;
    outregs->x.cflag = r.flags & 1; // carry flag
    return outregs->x.ax;
}

int real_int86(int intno, union REGS *inregs, union REGS *outregs)
{
    struct SREGS sregs;
    memset(&sregs, 0, sizeof(sregs)); 
    return real_int86x(intno, inregs, outregs, &sregs);
}

int real_intdos(union REGS *inregs, union REGS *outregs)
{
    return real_int86(0x21, inregs, outregs);
}

int real_intdosx(union REGS *inregs, union REGS *outregs, 

    struct SREGS *sregs)
{
    return real_int86x(0x21, inregs, outregs, sregs);
}

/**********************************************************************/

#pragma warn -rvl
WORD lar(WORD wSeg) // load access rights
{
    _asm lar ax, wSeg
    _asm jnz error
    _asm shr ax, 8
    _asm jmp short done; // value in AX
error:
    return 0;            // can't be a valid AR
done:;
}

void far *lin_to_real(DWORD lin_addr)
{
    if (lin_addr > 0x10FFEFL)   /* allow HMA pointers up to FFFF:FFFF */
        return NULL;            /* not accessible in real mode */
    else                    
    {
        WORD seg, ofs;
        seg = (lin_addr > 0x100000L) ? 0xFFFF : (lin_addr >> 4);
        ofs = lin_addr - ((DWORD) seg << 4L);
        return MK_FP(seg, ofs);
    }
}

void far *get_real_addr(void far *fp)   /* prot to real */
{
    DWORD base;
    if (lar(FP_SEG(fp)) == 0)
        return NULL;    /* not a valid pointer */
    base = GetSelectorBase(FP_SEG(fp));
    return lin_to_real(base + FP_OFF(fp));
}

/**********************************************************************/

BOOL alloc_real_seg(DWORD bytes, WORD *ppara, WORD *psel)
{
    DWORD dw = GlobalDosAlloc(bytes);
    if (! dw) return 0;
    *ppara = HIWORD(dw);
    *psel = LOWORD(dw);
    return 1;
}

BOOL free_real_seg(WORD sel)
{
    return (GlobalDosFree(sel) == 0);
}

/**********************************************************************/

#ifdef DPMI_APP
#pragma warn -rvl
WORD dpmi_alloc_selector(void)
{
    _asm xor ax, ax
    _asm mov cx, 1
    _asm int 31h
    _asm jc error
    _asm jmp short ok
error:
    return 0;
ok:;
    // return value in AX
}

WORD dpmi_free_selector(WORD sel)
{
    _asm mov ax, 1
    _asm mov bx, sel
    _asm int 31h
    _asm jc error
    _asm jmp short ok
error:
    return sel; // act like the Window function
ok:
    return 0;
}

WORD dpmi_set_selector_base(WORD sel, DWORD base)
{
    _asm mov ax, 7
    _asm mov bx, sel
    _asm mov cx, word ptr base+2
    _asm mov dx, word ptr base
    _asm int 31h
    _asm jc error
    _asm jmp short ok
error:
    return sel;
ok:
    return 0;
}

WORD dpmi_set_selector_limit(WORD sel, DWORD limit)
{
    _asm mov ax, 8
    _asm mov bx, sel
    _asm mov cx, word ptr limit+2
    _asm mov dx, word ptr limit
    _asm int 31h
    _asm jc error
    _asm jmp short ok
error:
    return sel;
ok:
    return 0;
}

DWORD dpmi_get_selector_base(WORD sel)
{
    _asm mov ax, 6
    _asm mov bx, sel
    _asm int 31h
    _asm jc error
    // return CX:DX in DX:AX
    _asm mov ax, dx
    _asm mov dx, cx
    _asm jmp short done
error:
    return 0;   // problem: instinguishable from 1-byte segment!
done:;
}

/* C interface to LSL (Load Segment Limit) instruction */
#pragma warn -rvl
WORD lsl(WORD sel)
{
    if (! sel) return 0;   /* workaround 386 bug: Hummel, p.471 */
    _asm lsl ax, sel
}

DWORD dpmi_get_selector_limit(WORD sel)
{
    // no DPMI function; use LSL; should use 32-bit LSL though!
    return (DWORD) lsl(sel);
}

DWORD dpmi_dos_alloc(DWORD bytes)
{
    DWORD retval;
    WORD paras = (bytes >> 4) + 1;
    _asm mov ax, 0100h
    _asm mov bx, paras
    _asm int 31h
    _asm jc error
    _asm mov word ptr retval, ax
    _asm mov word ptr retval+2, dx
    return retval;
error:
    return 0L;
}

WORD dpmi_dos_free(WORD sel)
{
    _asm mov ax, 0101h
    _asm mov dx, sel
    _asm int 31h
    _asm jc error
    return 0;
error:
    return sel;
}
#endif
