/* 
RANDRW.C
What are the chances that a random pointer would not GP fault?
Appears to be about a 1% chance that a random ptr in Win95 will trash memory
Andrew Schulman (76320.302@compuserve.com), March 1995

Using Microsoft C rand(); Win95:
1000000 attempts
875367 hits
87.54% hit rate
21 seconds
47619 attempts/second
41684 hits/second

Using my_rand(); Win95:
1000000 attempts
19630 hits
1.96% hit rate
422 seconds
2369 attempts/second
46 hits/second

Keeping list of writeable pointers encountered:

C1851EFC C1961639 C1885492 827E4BAF C0052BD1 C132D0EB 828887FC C209DF29 
C1FE33F4 00520A5A C1549390 82A50699 C10D7D0E 801E2D2E C026072B 000BCF36 
8017BE06 C1113A43 828DE806 829D4875 82971124 C1E22D1A C1308ABA 828AD65A 
C1A159E0 80376A3D C1F9A776 C3273C9C C125FAB1 C0FF8F98 C1C135F1 C22C694E 
...

WHAT THIS MEANS:

Program generated one million random pointers, and tried to write to
the pointer.  Of these, over 98% caused GP faults: This is exactly
what we want in a protected operating system.  However, almost
20,000 of the random memory writes succeeded.  The program was nice
and wrote a value identical to that already found at the address, so
no harm was done.

Win32s not as bad!:

1000000 attempts
5921 hits
0.59% hit rate
278 seconds
3597 attempts/second
21 hits/second

In Win32s, falls into two areas:
    00010B8F-0140F70E
    80011F3E-8871000E
Have to see what these correspond to.

-- Make -DANGER switch to write a different value.

-- Give complete instructions in output.

-- Make this RANDROUL?

-- Instead of hashtab to store all pointers, jump store count by MB

-- Simulate linked-list chains:  how long are randomly-generated
   list link chains (p = *p) before exception?

-- Check the random distribution.  

-- Put comment in code explaining why 69069 is a good number; also
   why intermediate floating point used.

-- Get quotes from OS textbooks and from MSDN CD-ROM re: protection
   from errant or malicious programs.
*/


#include <stdlib.h>
#include <stdio.h>
#include <limits.h>
#include <time.h>
#define WIN32_LEAN_AND_MEAN
#include "windows.h"

FILE *f;

#ifdef NOT_CONSOLE
void fail(const char *s) { MessageBox(0, s, "RANDRW", MB_OK); exit(1); }
#else
void fail(const char *s) { puts(s); exit(1); }
#endif

/**********************************************************************/

/* The rand() function in 32-bit Microsoft C, as in most (?) 32-bit
   C run-time libraries, has RAND_MAX = 0x7FFF (STDLIB.H).  This is
   too small.  We need even coverage of the entire range 0-UINT_MAX.
   This simple random-number generator (from George Marsaglia,
   "The Mathematics of Random Number Generators," Proceedings of
   Symposia in Applied Mathematics v. 46 (1992)) has a period of
   4.3 x 10^9 -- just what we need. */
unsigned seed;
unsigned my_srand(unsigned x) { seed = x; }
unsigned my_rand(void) { return (seed = (69069.0 * (seed + 1.0))); }

/**********************************************************************/

typedef struct _node {
    struct _node *next;
    unsigned p;
    } NODE;

void *allocate(size_t x, size_t y)
{
    void *p;
    if (! x) x = 1;
    if (! y) y = 1;
    if (((long) x * (long) y) > 0xFFFFUL)
        fail("internal error");
    if ((p = calloc(x, y)) == 0)
        fail("insufficient memory");
    return p;
}
    
NODE *mk_list(void)
{
    return (NODE *) 0;
}
    
void insert(NODE **list, unsigned p)
{
    NODE *node = (NODE *) allocate(1, sizeof(NODE));
    node->next = *list;
    *list = node;
    node->p = p;
}

unsigned find(NODE *list, unsigned p)
{
    while (list)
    {
        if (list->p == p)
            return list->p;
        else
            list = list->next;
    }
    return 0;   // didn't find:  needs to be element not in set!
}

void print_list(NODE *list)
{
    static int col = 0;
    while (list)
    {
        fprintf(f, "%08lX ", list->p);
        list = list->next;
        col++;
        if ((col % 8) == 0) fprintf(f, "\n");
    }
}

unsigned TAB_SIZE = 1021;

#define HASH(x)         ((x) % TAB_SIZE)

NODE **mk_hashtab(void)
{
    return (NODE **) allocate(TAB_SIZE, sizeof(NODE *));
}

void set(NODE **hashtab, unsigned p)
{
    insert(&hashtab[HASH(p)], p);
}

unsigned get(NODE **hashtab, unsigned p)
{
    return find(hashtab[HASH(p)], p);
}

void print_hashtab(NODE **hashtab)
{
    int i;
    for (i=0; i<TAB_SIZE; i++)
        print_list(hashtab[i]);
}

/**********************************************************************/

#ifdef NOT_CONSOLE
#define argc __argc
#define argv __argv
extern int __argc;
extern char **__argv;

int PASCAL WinMain(HANDLE hInstance, HANDLE hPrevInstance,
    LPSTR lpszCmdLine, int nCmdShow)
#else
main(int argc, char *argv[])
#endif

{
    NODE **tab = mk_hashtab();
    int iter = (argc < 2) ? 1000000 : atoi(argv[1]);
    int i, x, num_hits = 0, did_hit;
    unsigned *p;
    unsigned runtime;
    time_t t1, t2;
    
#ifdef NOT_CONSOLE
    f = fopen("randrw.log", "w");
#else
    f = stdout;
#endif
    
    time(&t1);
    my_srand(t1);
    for (i=0; i<iter; i++)
    {
        p = (unsigned *) my_rand();
        _try { did_hit = 1; x = *p; *p = x; }
        _except (EXCEPTION_ACCESS_VIOLATION) { did_hit = 0; }
        if (did_hit) { num_hits++; set(tab, (unsigned) p); }
    }
    time(&t2);
    runtime = t2 - t1;
    
    fprintf(f, "%u attempts\n", iter);
    fprintf(f, "%u hits\n", num_hits);
    fprintf(f, "%0.2f%% hit rate\n", (100.0 * num_hits) / iter);
    fprintf(f, "%u seconds\n", runtime);
    fprintf(f, "%u attempts/second\n", iter/runtime);
    fprintf(f, "%u hits/second\n", num_hits/runtime);
    fprintf(f, "\n");
    print_hashtab(tab);
    fprintf(f, "\n");
    fclose(f);
    return 0;
}

