/* 
RIEMANN.C
Uses the Riemann zeta function to output Ri(x), a very close approximation
to the number of primes <= x.
See Hans Riesel, "Prime Numbers and Computer Methods for Factorization",
Chapter 2 (Birkhauser, 1985).

Andrew Schulman
76320.302@compuserve.com

Uses stdio, WINIO, or MessageBox/sprintf for output.

To try to keep timing accurate, does not yield control to other apps
under Windows 3.1.

If compiled with -DSLOW, uses stupid (but valid) techniques for computing
the Riemann zeta function, and factorial.

With SLOW, uses a lot of floating-point arithmetic; 
results under Win32s are not good!

On 486dx/66 (built-in floating point):
    regular (not SLOW):
        16-bit:             0.11 seconds
        Win32s:             0.0 seconds
    SLOW:
        16-bit:             2.42 seconds
        Win32s:             6.0 seconds (why slower here???)

On 386sx/20 (no coprocessor; uses floating point emulator):
    regular (not SLOW):
        16-bit:             2.0 seconds
        Win32s:             7.0 seconds (why slower here???)
    SLOW:
        16-bit:             643.52 seconds
        Win32s:             3778 seconds (yikes!)
            
Incidentally, compiling with CL386 -Ox produces very odd results:
only one iteration in the do_riemann() loop is executed.  Compiling
with CL386 without the -Ox switch works fine.  And this only occurs
on the 386sx.  Weird!
*/

#define WINDOWS

#include <stdlib.h>
#include <stdio.h>      
#ifdef WINDOWS
#include "windows.h"
#else
#include <stdio.h>
#endif
#include <malloc.h>
#ifdef __BORLANDC__
#include <mem.h>
#endif
#include <math.h>
#include <float.h>
#include <limits.h>
#include <time.h>

#ifdef WINDOWS
void fail(char *s) { MessageBox(0, s, "RIEMANN", MB_OK); exit(1); }
#else
void fail(char *s) { puts(s); exit(1); }
#endif

#if 1
#define EQ(f1, f2)      ((f1)-(f2) < .5)    // ok for our purposes here
#else
// correct way to test for floating-point equality
#define EQ(f1, f2)      (fabs((f1) - (f2)) < DBL_EPSILON)
#endif

#ifndef SLOW
// table based on zeta with 1,000 iterations
// zeta(1) goes very, very slowly to infinity; the rest, even slower!
static double zetatab[] = {
    7.484471,   1.643934,   1.202056,   1.082323,   1.036928,   1.017343,   
    1.008349,   1.004077,   1.002008,   1.000995,   1.000494,   1.000246,   
    1.000123,   1.000061,   1.000031,   1.000015,   1.000008,   1.000004,   
    1.000002,   1.000001
    } ;
#endif

// zeta(x) = sum(i, 0, x, 1/(i^x))
double zeta(double x)
{
#ifdef SLOW
    double z;
    int i;
    for (i=1, z=0; i<1000; i++)         // can change #iter
        z += (1 / pow(i, x));
    return z;
#else
    if (x > 0 && x < 21) 
        return zetatab[(int) x-1];
    else if (x > 20)
        return 1;
    else
    {
        double z;
        int i;
        for (i=1, z=0; i<50; i++)           // can change #iter
        {
            z += (1 / pow(i, x));
            // printf("%u\t%f\n", i, z);    // to generate table
        }
        return z;
    }
#endif  
}

#define FACTTAB_MAX     60

// Factorial
// Build up table as we go along
// or, ifdef SLOW, just recompute every time
double fact(int x)
{
#ifdef SLOW
    double f;
    int i;
    if (x < 2) 
        return 1;
    for (i=2, f=1; i<=x; i++)
        f *= i;
    return f;
#else
    static int facttab_max = 0;
    static double *facttab = (double *) 0;
    double f;
    int i;
    
    if (! facttab_max)  // one-time initialization
    {
        if (! (facttab = (double *) malloc(FACTTAB_MAX * sizeof(double))))
            fail("Insufficient memory!");
        facttab[0] = facttab[1] = 1;
        facttab[2] = 2;
        facttab_max = 2;
    }
    
    if (x <= facttab_max)
        return facttab[x];
    else
    {
        for (i=facttab_max+1, f=facttab[facttab_max]; i<=x; i++)
        {
            f *= i;
            if (i < FACTTAB_MAX)
                facttab[++facttab_max] = f;
        }
        return f;
    }
#endif  
}

#define MAX_ITER    200

double riemann(double x)
{
    double last = 0;
    double r;
    int i;
    for (i=1, r=1; i<MAX_ITER; i++)
    {
        r += (pow(log(x), i) / (i * zeta(i + 1) * fact(i)));
        if (EQ(r, last))    // converges pretty rapidly
            break;
        last = r;
        // printf("%u\t%f\n", i, r);
    }
    return r;
}

static double primes = 0;

float runtime(double (*f)(void))
{
    clock_t t = clock();
    primes = (*f)();
    return (clock() - t) / CLOCKS_PER_SEC;
}

double do_riemann(void)
{
    double x, rx;
    for (x=10; x<=100000000000.0; x *= 10)
    {
        rx = riemann(x);
        // printf("%12.0f\t%12.0f\t%.2f%%\n", x, rx, 100 * (rx / x));
    }
    return rx;
}

#ifdef WINDOWS
int PASCAL WinMain(HANDLE hInstance, HANDLE hPrevInstance, 
    LPSTR lpCmdLine, int nCmdShow)
{
    char buf[80];
    double t = runtime(do_riemann);
    sprintf(buf, "About %.0f primes < 100,000,000,000\n%.2f seconds", 
        primes, t);
    MessageBox(0, buf, "RIEMANN", MB_OK);
    return 0;
}
#else
int main(void)
{
    printf("%.2f seconds\n", runtime(do_riemann));
    return 0;
}
#endif
