/* 
RW.C -- Win32 read/write memory app; demonstrate Win32 non-robustness
Andrew Schulman (76320.302@compuserve.com), March 1995

This is a Win32 Console application; combine with 32-bit Microsoft C++
(cl rw.c)

Windows 95 maps a significant part of lower DOS memory into each
Win32 process.  Empirically, it seems to be the range of a little
over one megabyte from linear address 1000h to 124000h.  This
includes all conventional memory starting at 0100:000, all upper
memory blocks (UMBs), and the high memory area (HMA).  This is
*global* memory visible to, and writeable by, all Win32 processes.
In a typical Windows 95 configuration, it includes items such as the
following (as reported by MEM /D):

   0020F             56,240   (55K)  IO           System Data
                      1,024    (1K)               Relocated EBIOS data 
                        688    (1K)     SETVERXX  Installed Device=SETVER  
                      1,088    (1K)     XMSXXXX0  Installed Device=HIMEM   
                     22,016   (22K)     NECCD0    Installed Device=NEC_IDE 
                      1,504    (1K)     DBLSYSH$  Installed Device=DRVSPACE
                      2,432    (2K)     DblBuff$  Installed Device=DBLBUFF 
                      2,800    (3K)     IFS$HLP$  Installed Device=IFSHLP  
                        544    (1K)               Sector buffer 
                        304    (0K)               Block device tables 
                      1,488    (1K)               FILES=30
                        256    (0K)               FCBS=4
                     15,968   (16K)               BUFFERS=30
                      2,816    (3K)               LASTDRIVE=`
                      3,072    (3K)               STACKS=9,256
   00FCA                 80    (0K)  MSDOS        System Program
   00FCF                 32    (0K)  WIN          Data
   00FD1                 80    (0K)  vmm32        Data
   00FD6                 32    (0K)  vmm32        Data
   00FD8                112    (0K)  vmm32        Data
   00FDF              4,688    (5K)  DOSKEY       Program
   01104                256    (0K)  WIN          Environment
   01114              3,136    (3K)  WIN          Program
   011D8                640    (1K)  vmm32        Program
   01200                272    (0K)  COMMAND      Data
   01211              8,688    (8K)  COMMAND      Program
   01430              2,080    (2K)  COMMAND      Environment
       
It also includes many important interrupt vectors, such as these
displayed by my INTVECT program:

INT 13h   FD96:25CD              arpl -- Windows V86 breakpoint
INT 14h   F000:E739              jmp F000:F8FA
INT 15h   027E:0240   XMSXXXX0   
INT 17h   F000:EFD2              jmp F000:F85E
INT 18h   F000:E000              jmp F000:E005
INT 19h   0FCB:002F   IFS$HLP$   jmp 0070:036D
INT 1Ah   F000:FE6E              
INT 1Bh   091C:0604   IFS$HLP$   
INT 21h   091C:04A0   IFS$HLP$   

The last one, INT 21h, is quite important:  it's the DOS handler.
Overwriting this with garbage will crash Windows 95 on the next INT 21h 
call.  

Fortunately, some of these addresses are in the ROM BIOS and thus not
writeable.  This is particularly important for the first item shown
here, with a Windows V86 breakpoint (ARPL).  But this also means that
the SystemROMBreakPoint=off switch (see *Unauthorized Windows 95*,
pp. 256-257) is especially dangerous in Windows 95.

If you want to see which other areas of a Win32 process' address space
are read/writeable, you can specify a different starting address and
number of bytes on the RW command line.  Naturally, in many cases Win95
will detect the illegal read/write and shut down the app.  However, if
you see the "Successfully read/wrote" message at the end, it means the
entire range specified was readable and writeable.

The bottom line:

(1) Even if you only run Win32 applications, Win95 has an enormous
vulnerability to ill-behaved applications.  

(2) In this respect, Win32 applications are *more* dangerous than
Win16 apps, because a Win16 app must *explicitly* ask to read this
DOS memory, whereas for a Win32 app it's *automatically* mapped in.  

(3) Not surprisingly, a DOS application can write to this memory
(this is easily demonstrated with DEBUG, as both Rob Hummel and
Richard Smith have pointed out).  What's perhaps surprising is that
overwriting DOS from a DOS app will affect all Win16 and Win32 apps.

While RW is running, you can press V to toggle a verbose mode that
prints out any interesting-looking strings encountered in memory, or 
Q to quit.
*/

#include <stdlib.h>
#include <stdio.h>
#include <ctype.h>
#include <conio.h>

main(int argc, char *argv[])
{
    unsigned bytes;
    unsigned *p, *orig;
    unsigned *top;
    unsigned u;
    int verbose = 0;
    int i;

    p = (unsigned *) 0x1000;
    bytes = 0x123000;
    if (argc > 1) sscanf(argv[1], "%X", &p);
    if (argc > 2) sscanf(argv[2], "%X", &bytes);
    top = p + (bytes / sizeof(unsigned));
    printf("Try to read/write %0lXh (%u) bytes from %08lXh to %08lXh\n",
        bytes, bytes, p, top);
    
    for (orig = p; p < top; p++)
    {
        u = *p;     // read
        // we could do whatever we wanted to "u" in here!
        *p = u;     // write
            
        if (verbose) 
        {
            char *s = (char *) p;
            int i;
            #define MAYBE_PUT(c) if (isalnum(c)) putchar(c)
            for (i=0; i<4; i++) MAYBE_PUT(s[i]);
        }
            
        if (kbhit())
        {
        top:
            switch (getch())
            {
                case 'q' : case 'Q' : goto done;
                case 'v' : case 'V' : verbose = ! verbose; break;
                case ' '            : getch(); break;
                default             : 
                    printf("\nPress V to toggle verbose\n");
                    printf("Press Q to quit\n");
                    goto top;
            }
        }
    }

    done:
    printf("\nSuccessfully read/wrote %0lXh (%u) bytes "
           "from %08lXh to %08lXh\n",
        (p-orig)*sizeof(unsigned), (p-orig)*sizeof(unsigned), 
        orig, p);
    return 0;
}
