/*
sets.c

Andrew Schulman
13 March 1988

Assignment:
    Write a program to read in a positive integer n and two sets A and B,
    where U = {1,2,3,...,n}, and do the following:
        i)      Print their bit representations.
        Use the bit representations to find the following sets and
        their cardinalities:
        ii)     A u B       
        iii)    A n B
        iv)     A'
        v)      A - B
        vi)     A @ B
        vii)    A x B
    Test data:
        n = 8
        A = {1,2,3,5}
        B = {1,3,4,5,6}
            
Notes:
    1)  C does not have a built-in SET type, but small sets (with up to
        32 members) can be represented with an unsigned long.  If a bit is
        set (1), the corresponding item is in the set; if a bit is
        clear (0), the corresponding item is not in the set.  The bitwise
        The bitwise operators are used to implement the set operators:
            a)  set union is | (bitwise or)
            b)  set intersection is & (bitwise and)
            c)  set complement is ~ (bitwise negation)
            d)  set symmetric difference is ^ (bitwise xor)
            e)  set difference:  A-B is (A & ~B)
            f)  cartesian product:  this is trickier since our
                implementation of sets does not include sets of sets.
            
    2)  This code is based somewhat on the set package in Samuel P.
        Harbison and Guy Steele, _C: A Reference Manual_, Second Edition,
        pp.174-179.  See especially the unused code (it works, but it's
        weird!) in cardinality().
            
    3)  The lack of a built-in set type is one of C's disabilities
        compared with Modula-2 or Pascal.  C's enum type is only
        superficially similar to Modula-2 or Pascal sets.  
            
    4)  In contrast to this code's heavy use of preprocessor macros,
        a better implementation would use C++ operator overloading.
            
    5)  A more rigorous implementation of sets can be found in
        the C source code for Gary Levin's ISETL language.
            
    6)  The assignment calls for the first member of U to be 1, not 0.
        This means that the rightmost bit must be set with (1L << (i-1)).
            
    7)  Input for the program is entered on the DOS command-line with
        commas separating n, the members of A, and the members of B.
        The command-line for the test data is:
            C:\HOMEWORK>sets 8 , 1 2 3 5 , 1 3 4 5 6
        Empty sets can be entered like so:
            C:\HOMEWORK>sets 8 , , 3 4 5
                
    8)  The program does almost no error checking.
        
    9)  Compile with:
            cl sets.c
*/

#include <stdlib.h>
#include <stdio.h>

typedef unsigned long SET;      /* the key to the whole thing */

#define FIRST_MEMBER            1

#define LAST_MEMBER             card_U      /* input at run-time */

#define NULL_SET                ((SET) 0)   

#define UNION(a,b)              ((a) | (b))

#define INTERSECTION(a,b)       ((a) & (b))

#define COMPLEMENT(a)           (~(a))

#define DIFF(a,b)               INTERSECTION(a, COMPLEMENT(b))

#ifdef VERBOSE
#define SYMMETRIC_DIFF(a,b)     UNION(DIFF(a,b), DIFF(b,a))
#define XOR(a,b)                (a & ~b) | (b & ~a)
#else
#define SYMMETRIC_DIFF(a,b)     ((a) ^ (b))
#endif

#define SINGLETON(i)            (((SET) 1) << (i-1))

#define MEMBER(a,i)             INTERSECTION(a, SINGLETON(i))

#define MAKE_MEMBER(a,i)        ((a) |= SINGLETON(i))

unsigned cartesian_product(SET a, SET b);
unsigned cardinality(SET a);
void bitstring(SET a);
void print(SET a);
void show(char *msg, SET a);

static unsigned card_U;

main(int argc, char *argv[])
{
    SET a = NULL_SET;
    SET b = NULL_SET;
    unsigned card_prod;
    register int i;
    register int j;

    /* get the size of the "universe of discourse" */
    card_U = atoi(argv[1]);
    if (card_U < 1 || card_U > 32)
    {
        printf("n(U) must be 1..%u\n", sizeof(SET)*8);
        exit(1);
    }
    
    /* get the two sets */
    for (i=0, j=3; i<card_U && *(argv[j]) != ','; i++, j++)
        MAKE_MEMBER(a, atoi(argv[j]));
    j++;
    for (i=0; i<card_U && j < argc; i++, j++)
        MAKE_MEMBER(b, atoi(argv[j]));
    
    /* print the sets, their bit representations, and their cardinalities */
    show("A", a);
    show("B", b);
    
    /* use bit representations to perform various operations */
    show("A u B", UNION(a,b));
    show("A n B", INTERSECTION(a,b));
    show("A'", COMPLEMENT(a));
    show("A - B", DIFF(a,b));
    show("A @ B", SYMMETRIC_DIFF(a,b));
    printf("A x B:  "); card_prod = cartesian_product(a,b);
    printf("n(A x B): %u\n", card_prod);
}

void bitstring(SET a)
{
    register int i;
    
    for (i=LAST_MEMBER; i>=FIRST_MEMBER; i--)
        putchar(MEMBER(a, i) ? '1' : '0');
    putchar('\n');
}

/* prints out Cartesian product and returns its cardinality */
unsigned cartesian_product(SET a, SET b)
{
    int first_time = 1; 
    unsigned card = 0;
    register int i;
    register int j;
    
    printf("{");
    for (i=FIRST_MEMBER; i<=LAST_MEMBER; i++)
        if (MEMBER(a, i))
            for (j=FIRST_MEMBER; j<=LAST_MEMBER; j++)
                if (MEMBER(b, j))
                {
                    printf("%s(%u,%u)", (first_time) ? "" : ", ", i, j);
                    first_time = 0;
                    card++;
                }
    printf("}\n");
    return card;
}

unsigned cardinality(SET a)
{
    unsigned c = 0;
#ifdef TRICKY
    while (a != NULL_SET)
    {
        a ^= (a & -a);
        /* printf("%u\tA=%s\t", c, bitstring(a)); print(a); */
        c++;
    }
#else
    register int i;
    for (i=FIRST_MEMBER; i<=LAST_MEMBER; i++)
        if (MEMBER(a, i))
            c++;
#endif      
    return c;
}

void print(SET a)
{
    int first_time = 1;
    register int i;

    printf("{");
    for (i=FIRST_MEMBER; i<=LAST_MEMBER; i++)
        if (MEMBER(a, i))
        {
            printf("%s%u", (first_time) ? "" : ", ", i);
            first_time = 0;
        }
    printf("}\n");
}

void show(char *msg, SET a)
{
    printf("%s:\t", msg); bitstring(a);
    printf("%s:\t", msg); print(a);
    printf("n(%s): %u\n\n", msg, cardinality(a));
}
    
/*
output:

C:\HOMEWORK>sets 8 , 1 2 3 5 , 1 3 4 5 6
A:  00010111
A:  {1, 2, 3, 5}
n(A): 4

B:  00111101
B:  {1, 3, 4, 5, 6}
n(B): 5

A u B:  00111111
A u B:  {1, 2, 3, 4, 5, 6}
n(A u B): 6

A n B:  00010101
A n B:  {1, 3, 5}
n(A n B): 3

A': 11101000
A': {4, 6, 7, 8}
n(A'): 4

A - B:  00000010
A - B:  {2}
n(A - B): 1

A @ B:  00101010
A @ B:  {2, 4, 6}
n(A @ B): 3

A x B:  {(1,1), (1,3), (1,4), (1,5), (1,6), (2,1), (2,3), (2,4), (2,5), (2,6),
(3,1), (3,3), (3,4), (3,5), (3,6), (5,1), (5,3), (5,4), (5,5), (5,6)}
n(A x B): 20
*/
