/*
    SYSTIMER.C -- Demonstrates CreateSystemTimer and KillSystemTimer

    From Chapter 9 of "Undocumented Windows" (Addison-Wesley 1992)
    by Andrew Schulman, Dave Maxey and Matt Pietrek

    Build using: WINIOBC SYSTIMER (for Borland C++ v3.00)
                 WINIOMS SYSTIMER (for Microsoft C/SDK)

	usage:
		systimer [loops] [rate] [any third arg disables tight loop]
*/

#include <windows.h> 
#include <stdlib.h> 
#include "winio.h" 
#include "wmhandlr.h"

#define LOOPS  200
#define RATE   100

#define INQSYS_TIMERRES     0

WORD FAR volatile gTicks = 0;
WORD FAR volatile gGotTick = 0;
WORD FAR giRate;

/*
    NOTE:  Because it will be called directly, at interrupt time,
    from SYSTEM's INT 8 handler, the callback function installed
    with CreateSystemTimer() *MUST* be located in a FIXED segment.
    For this example program, we've made the entire code segment
    fixed by changing the .DEF file statement from the normal
    CODE PRELOAD MOVEABLE DISCARDABLE to simply CODE PRELOAD (i.e.,
    non-moveable, non-discardable).  In a genuine application, the
    callback should be located in its own fixed segment.  On the
    other hand, the callback does *NOT* have to be located in a
    DLL, even though Microsoft's documentation suggests that this is
    necessary for Windows interrupt handlers.
*/
void FAR _export TimerFunc(void)
{
    gTicks++;
    gGotTick = 1;
    if (giRate >= 500)
        MessageBeep(0);  // this just happens to work
    // PostMessage is guaranteed to work at interrupt time
    // other simple functions:  GetCurrentTask, etc.
}

static WORD hTimer = 0;
static FARPROC fpTimerProc = 0;
static WORD (FAR PASCAL *KillSystemTimer)(WORD hTimer);

void cleanup(void)
{
    if (hTimer)
    {
        if (KillSystemTimer(hTimer) != 0)   // 0 = success
            fail("Couldn't kill timer");
        hTimer = 0;
    }
    if (fpTimerProc)
    {
        FreeProcInstance(fpTimerProc);
        fpTimerProc = 0;
    }
}
    
// just in case they close window while system timer is installed
void on_close(HWND hwnd) { cleanup(); }

int main(int argc, char *argv[])
{
    WORD (FAR PASCAL *CreateSystemTimer)(WORD wRate, FARPROC fpCallback);
    DWORD (FAR PASCAL *InquireSystem)(WORD wFlag, WORD wOption);
    DWORD dwTimerRes;
    DWORD t1, t2;
    WORD hSystem;
    BOOL bTightLoop;
    int iLoopMax;
    int iLoop, iLoopLoop;
    
    winio_about("SYSTIMER"
        "\nDemonstrates CreateSystemTimer and KillSystemTimer"
        "\n\nFrom Chapter 9 of"
        "\n\"Undocumented Windows\" (Addison-Wesley, 1992)"
        "\nby Andrew Schulman, David Maxey and Matt Pietrek"
        );
    
    iLoopMax = (argc < 2) ? LOOPS : atoi(argv[1]);
    giRate = (argc < 3) ? RATE : atoi(argv[2]);
    bTightLoop = (argc < 4) ? 1 : 0;

    /* dynamically link to undocumented SYSTEM functions */
    hSystem = GetModuleHandle("SYSTEM");
    CreateSystemTimer = GetProcAddress(hSystem, "CREATESYSTEMTIMER");
    KillSystemTimer = GetProcAddress(hSystem, "KILLSYSTEMTIMER");
    InquireSystem = GetProcAddress(hSystem, "INQUIRESYSTEM");

    if (CreateSystemTimer && KillSystemTimer && InquireSystem)
        puts("Installing asynchronous timer with CreateSystemTimer...");
    else
        fail("Can't find CreateSystemTimer and KillSystemTimer");
    
    dwTimerRes = InquireSystem(INQSYS_TIMERRES, 0);
    if (giRate < (dwTimerRes / 1000))
        printf("Warning: Lowest rate is %u\n", dwTimerRes / 1000);

    gTicks = 0;
    
    /* install the asynchronous timer handler */
    fpTimerProc = MakeProcInstance((FARPROC) TimerFunc, __hInst);
    if ((hTimer = CreateSystemTimer(giRate, fpTimerProc)) == 0)
        fail("CreateSystemTimer failed");
    
    winio_onclose(__hMainWnd, on_close);
    
    t1 = GetCurrentTime();

    if (bTightLoop)
    {
        /*
            NOTE!  The following is a tight loop, with no calls to
            GetMessage or PeekMessage.  If our TimerFunc gets called,
            it means system timers really are asynchronous.
        */

        for (iLoop = 0; iLoop < iLoopMax; iLoop++)
            for (iLoopLoop = 0; iLoopLoop != -1; iLoopLoop++)
                    ;
    }
    else
    {
        /* If we weren't trying to show that the timer function is
           called asynchronously, we would probably use system timers
           in the following way */

        for (iLoop = 0; iLoop < iLoopMax; iLoop++)
            for (iLoopLoop = 0; iLoopLoop != -1; iLoopLoop++)
                if (gGotTick != 0)  // check volatile variable
                {
                    printf("%u\n", gTicks); // printf will yield
                    gGotTick = 0;
                }
    }
    
    t2 = GetCurrentTime();
    
    cleanup();

    printf("System timer handle: %04xh\n", hTimer);
    printf("SetSystemTimer rate: %u\n", giRate);
    printf("GetCurrentTime elapsed: %lu\n", t2 - t1);
    printf("Calls to handler: %u\n", gTicks);
    if (gTicks != 0)
        printf("Actual rate: %lu\n", (t2 - t1) / gTicks);

    return 0;
}
