/* 
taut2.c

Write a program to determine if (p->q)^(q->r)->(p->r) is a tautology.
Display the truth table.
Note: p, q, and r are arbitrary propositions.

This program differs from tautology.c: instead of representing the
Boolean operators AND and IMPLICATION as code, here they are represented
as data.  A Boolean operator is just a little look-up table.  See
the function f().  All 16 binary Boolean operators (known in graphics
programming as "raster ops") are given, just so the code can be
re-used for solving a different proposition at a later time.

Note that the truth table is ordered from F to T instead of from
T to F.  It's ordered as in Boolean algebra rather than as in 
propositional logic truth tables, because it seems to make more
sense to do it this way on a computer.

Andrew Schulman
3 February 1989

compile with: cl taut2.c
*/

#include <stdio.h>

#pragma pack(1)

#define FALSE               0
#define TRUE                1

#define TF(x)               ((x) ? 'T' : 'F')

typedef unsigned BOOL;

typedef struct {
    unsigned char t[4];
    } BOOLOP;
    
BOOLOP bool[16] = {
    0, 0, 0, 0,         /* 0 */
    0, 0, 0, 1,         /* 1 */
    0, 0, 1, 0,         /* 2 */
    0, 0, 1, 1,         /* 3 */
    0, 1, 0, 0,         /* 4 */
    0, 1, 0, 1,         /* 5 */
    0, 1, 1, 0,         /* 6 */
    0, 1, 1, 1,         /* 7 */
    1, 0, 0, 0,         /* 8 */
    1, 0, 0, 1,         /* 9 */
    1, 0, 1, 0,         /* 10 */
    1, 0, 1, 1,         /* 11 */
    1, 1, 0, 0,         /* 12 */
    1, 1, 0, 1,         /* 13 */
    1, 1, 1, 0,         /* 14 */
    1, 1, 1, 1,         /* 15 */
    };

#define FFFF    bool[0]             /* contradiction */
#define FFFT    bool[1]             /* AND: p ^ q */
#define FFTF    bool[2]
#define FFTT    bool[3]             /* p */
#define FTFF    bool[4]
#define FTFT    bool[5]             /* q */
#define FTTF    bool[6]             /* XOR: ~(p=q) */
#define FTTT    bool[7]             /* OR: p V q */
/* --------------- mirror image ---------------- */
#define TFFF    bool[8]             /* NOR: ~(p V q) */
#define TFFT    bool[9]             /* biconditional: p=q */
#define TFTF    bool[10]            /* NOT q */
#define TFTT    bool[11]            /* converse: q->p */
#define TTFF    bool[12]            /* NOT p */
#define TTFT    bool[13]            /* implication: p->q */
#define TTTF    bool[14]            /* NAND, Sheffer stroke: ~(p ^ q) */
#define TTTT    bool[15]            /* tautology */

/* of the 16 "raster ops", we're only using two */
#define IMPLIC  TTFT
#define AND     FFFT

BOOL f(BOOL x, BOOLOP op, BOOL y);
void print_truth_row(BOOL p, BOOL q, BOOL r, BOOL s);

main()
{
    BOOL p, q, r, s;
    BOOL tautology = TRUE;      /* may be set to FALSE within loop */

    printf("p \tq \tr \t(p->q)^(q->r)->(p->r)\n");
    printf("- \t- \t- \t---------------------\n");
    
    for (p = FALSE; p <= TRUE; p++)
        for (q = FALSE; q <= TRUE; q++)
            for (r = FALSE; r <= TRUE; r++)
            {   
                s = f(  f(  f(p,IMPLIC,q),      /* (p->q) */
                            AND,                /* ^      */
                            f(q,IMPLIC,r)       /* (q->r) */
                            ),
                        IMPLIC,                 /* ->     */
                        f(p,IMPLIC,r));         /* (p->r) */
                if (s == FALSE)
                    tautology = FALSE;
                print_truth_row(p, q, r, s);
            }
    
    puts(tautology ? "\nTautology" : "\nNot tautology");
}

/*
-- a Boolean operator is actually just a little look-up table!
-- for example, if (p == TRUE) and (q == FALSE), then (1 << 1) + 0 = 2,
   so we return op.t[2]. For the IMPLICATION operator this would yield
   FALSE.  
-- As another example, if (p == TRUE) and (q == TRUE), then (1 << 1) + 1 = 3,
   so we return op.t[3]. For the AND operator this would yield TRUE.
*/     
BOOL f(BOOL p, BOOLOP op, BOOL q)
{
    return op.t[(p << 1) + q];
}

void print_truth_row(BOOL p, BOOL q, BOOL r, BOOL s)
{
    printf("%c \t%c \t%c \t\t%c\n", TF(p), TF(q), TF(r), TF(s));
}

/*
output:

p   q   r   (p->q)^(q->r)->(p->r)
-   -   -   ---------------------
F   F   F       T
F   F   T       T
F   T   F       T
F   T   T       T
T   F   F       T
T   F   T       T
T   T   F       T
T   T   T       T

Tautology
*/

