/*
V86TEST.C -- take over INT 21h and INT 2Fh, count calls in V86 mode
from Schulman, Unauthorized Windows, 1994

Some of the complexity here is artificial, so that V86TEST could be 
tested with programs other than Windows, such as FAKEWIN.

Can use with WV86TEST

bcc -2 -P- v86test.c
*/

#include <stdlib.h>
#include <stddef.h>
#include <stdio.h>
#include <string.h>
#include <ctype.h>
#include <process.h>
#include <time.h>
#include <dos.h>

typedef unsigned short WORD;
typedef unsigned long DWORD;

#define VM_MAX      8
#define VM_OTHER    VM_MAX

#define GET_STATS   0xFFFF
#define SIGNATURE   "V86TEST"

#define VXD_MAX     0x100
#define VXD_OTHER   VXD_MAX

#if 0
struct {
    WORD vxd_id;
    DWORD num_calls;
    } VXDCALLS;

VXDCALLS vxdcalls[100] = {0};
int num_vxds = 0;
#endif

#pragma pack(1)

typedef struct {
#ifdef __TURBOC__
    WORD bp,di,si,ds,es,dx,cx,bx,ax;
#else
    WORD es,ds,di,si,bp,sp,bx,dx,cx,ax;     /* same as PUSHA */
#endif
    WORD ip,cs,flags;
    } REG_PARAMS;

void interrupt far int21(REG_PARAMS r);
void interrupt far int2f(REG_PARAMS r);

void (interrupt far *old)();
void (interrupt far *old_2F)();

#define WIN_INIT_NOTIFY     0x1605
#define WIN_INIT_COMPLETE   0x1608
#define WIN_EXIT_BEGIN      0x1609
#define WIN_EXIT_NOTIFY     0x1606

typedef enum { 
    NO_WIN, 
    WIN_INIT_BEGIN,             // got WIN_INIT_NOTIFY
    WIN_INIT_DONE,              // got WIN_INIT_COMPLETE
    WIN_FINI_BEGIN,             // got WIN_EXIT_BEGIN
    WIN_FINI_DONE,              // got WIN_EXIT_NOTIFY
    NUM_STATES } STATE;

static char *state_str[NUM_STATES] = {
    "Before Windows started", 
    "During Windows init",      // got WIN_INIT_NOTIFY
    "While Windows running",    // got WIN_INIT_COMPLETE
    "During Windows exit",      // got WIN_EXIT_BEGIN
    "After Windows exited",     // got WIN_EXIT_NOTIFY
    } ;

typedef struct {
    char signature[8];
    DWORD calls[NUM_STATES], v86_calls[NUM_STATES];
    DWORD iopl_count[4];
    DWORD vm[VM_MAX+1];
    DWORD int21[0x100];
    DWORD int2f[0x100], int2f16[0x100], int2f1607[VXD_OTHER+1];
    time_t start, end;
    } STATS;
    
static STATE state = NO_WIN;
static STATS *stats;
static STATS far *fpstats;
static int not_verbose = 1;
static int not_filter = 1;
static int v86_okay = 0;

char *usage = "usage: v86test [-okv86] [-filter | -verbose]"
    " [-query | -clear | win] <args...>";
        
#define PUT(s)  { fputs(s, stderr); fputs("\n", stderr); }
#define FAIL(s) { PUT(s); exit(1); }

/**********************************************************************/

int win3e(void) // is Windows Enhanced mode running?
{
    int maj = 0;
    _asm mov ax, 1600h
    _asm int 2fh
    _asm mov byte ptr maj, al
    return (maj && (maj != 0x80));
}

int pe(void)    // is processor Protect Enable bit set?
{
    _asm smsw ax
    _asm and ax, 1
    // retval in AX
}

int iopl(void)  // get I/O Protection Level from flags
{
    _asm pushf
    _asm pop ax
    _asm shr ax, 12
    _asm and ax, 3
    // retval in AX
}

int vmid(void)  // Get Windows Virtual Machine ID. If Windows Enh mode 
{               // is running, V86TEST will never see this 2f/1683 call!
                // And if we tried calling it with PUSHF/CALLF rather
                // than with INT, it would get the wrong results!
    _asm mov ax, 1683h
    _asm int 2fh
    _asm mov ax, bx
    // retval in AX
}

/**********************************************************************/

void display_results(STATS far *fp2)
{
    STATS *fp;
    DWORD elapsed;
    time_t start, end;
    int i;
    
    if (! (fp = malloc(sizeof(STATS))))
        FAIL("Insufficient memory");
    // copy over so stats don't change while reading them
    _fmemcpy(fp, fp2, sizeof(STATS));
    
    printf("\n");
    for (i=NO_WIN; i<NUM_STATES; i++)
        printf("%s:\t%lu INT 21/2F calls, %lu in V86 mode\n",
            state_str[i], fp->calls[i], fp->v86_calls[i]);

    printf("\nWhile Windows running:\n");
    if (fp->end) end = fp->end; else time(&end);
    if (fp->start) start = fp->start; else time(&start);
    if ((elapsed = end - start) != 0)
    {
        printf("Windows active for %lu seconds\n", elapsed);
        printf("%lu INT 21/2F calls/second\n", 
            fp->calls[WIN_INIT_DONE] / elapsed);
    }

    for (i=0; i<4; i++)
        if (fp->iopl_count[i])
            printf("IOPL=%d -- %lu calls\n", i, fp->iopl_count[i]);

    for (i=0; i<VM_MAX; i++)
        if (fp->vm[i])
            printf("VM #%d -- %lu calls\n", i, fp->vm[i]);
    if (fp->vm[VM_OTHER])
        printf("VM > #%d -- %lu calls\n", VM_MAX, fp->vm[VM_OTHER]);

    // put before verbose check: always show 2f/16 calls
    printf("\nINT 2Fh AH=16h calls seen by V86TEST:\n");
    for (i=0; i<0x100; i++)
        if (fp->int2f16[i])
            printf("%02X: %lu\t", i, fp->int2f16[i]);
    printf("\n");

    if (not_verbose)
        return;

    // get following by running V86TEST -VERBOSE
    printf("\nINT 21h calls:\n");
    for (i=0; i<0x100; i++)
        if (fp->int21[i])
            printf("%02X: %lu\t", i, fp->int21[i]);

    printf("\n\nINT 2Fh calls:\n");
    for (i=0; i<0x100; i++)
        if (fp->int2f[i])
            printf("%02X: %lu\t", i, fp->int2f[i]);

    printf("\n\nINT 2Fh AX=1607h calls:\n");
    for (i=0; i<VXD_MAX; i++)
        if (fp->int2f1607[i])
            printf("%02X: %lu\t", i, fp->int2f1607[i]);
    if (fp->int2f1607[VXD_OTHER])
        printf("VxD>#%04X: %lu\n", VXD_MAX, fp->int2f1607[VXD_OTHER]);
    printf("\n\n");
    
    free(fp);
}

STATS far *get_stats(void)  // call resident copy of V86TEST
{
    STATS far *fp;
    _asm mov ax, GET_STATS
    _asm int 2fh
    _asm mov word ptr fp+2, es
    _asm mov word ptr fp, bx
    return (_fstrcmp(fp->signature, SIGNATURE) == 0) ? fp : 0;
}

main(int argc, char *argv[])
{
    int i;
    
    PUT("V86TEST -- Test effect of Windows on software loaded before it");
    PUT("From \"Unauthorized Windows\" (IDG Books, 1994)");
    PUT("Copyright (c) 1994 Andrew Schulman.  All rights reserved.\n");
    
    /* look for command-line switches */
    while (argv[1][0] == '-')
    {
        STATS far *fp;
        switch (toupper(argv[1][1]))
        {
            case 'F' :  // filter
                not_filter = 0;
                break;
            case 'O' :  // okv86
                v86_okay = 1;
                break;
            case 'V' :  // verbose
                not_verbose = 0;
                break;
            case 'Q' :  // query
                if (! (fp = get_stats()))
                    FAIL("Can't get V86TEST statistics");
                display_results(fp); 
                exit(0);
            case 'C' :  // clear
                if (! (fp = get_stats()))
                    FAIL("Can't clear V86TEST statistics");
                _fmemset(&fp->calls, 0, 
                    sizeof(STATS) - offsetof(STATS, calls));
                exit(0);
            default:   
                FAIL(usage);
        }
        argv++; argc--;
    }

ok:     
    if (argc < 2) FAIL(usage);
    
    if (win3e()) FAIL("Already running under Windows Enhanced mode\n"
                      "Exit Windows and try again");

    if (! v86_okay)
    if (pe()) FAIL("Already in V86 mode -- test would be pointless\n"
                   "Remove 386 memory manager from CONFIG.SYS and reboot");

    if ((stats = calloc(1, sizeof(STATS))) == 0)
        FAIL("Insufficient memory");
    fpstats = (STATS far *) stats;
    strcpy(stats->signature, SIGNATURE);
    
    /* hook INT 21h and INT 2Fh */
    old = _dos_getvect(0x21);
    _dos_setvect(0x21, int21);
    old_2F = _dos_getvect(0x2F);
    _dos_setvect(0x2f, int2f);

    /* run command */
    spawnvp(P_WAIT, argv[1], &argv[1]);
    
    /* unhook INT 21h, 2Fh */
    _dos_setvect(0x2F, old_2F);
    _dos_setvect(0x21, old);
    
    display_results(fpstats);
    return 0;
}

/**********************************************************************/

void check_state(int intno, int ah, int al, int bx)
{
    stats->calls[state]++;

    if (pe())
        stats->v86_calls[state]++;

    if (state == WIN_INIT_DONE)
    {
        /* The following is a little confusing. V86TEST is normally able
           to make Windows INT 2Fh calls (such as vmid()) from inside an
           INT 2Fh handler, without causing an endless loop, because VMM
           gets these INT 2Fh calls and doesn't reflect them back to V86
           mode. Thus, V86TEST mode's INT 2Fh handler never sees
           calls such as 2F/1680 and 2F/1683. However, it's useful to
           test V86TEST with programs other than Windows, such as FAKEWIN.
           In this case, the 2F/168x calls would wind up right back at
           V86TEST, causing an endless loop if called from inside the
           INT 2Fh handler. So if genuine Windows VMM isn't running (as
           determined by the 2F/1680 call), V86TEST doesn't call 2F/1683.
           2F/1680 is called just once; logging is turned off for that one 
           call to prevent an endless loop if something other than Windows 
           VMM is running. */

        int cur_vm;
        
        static int logging = 1;
        static int is_win = 0xff;
        if (logging == 0) return;
        else if (is_win == 0xff) // one-time init
        {
            logging = 0;
            is_win = win3e();
            logging = 1;
        }
                
        cur_vm = (is_win) ? vmid() : 0;

        stats->vm[(cur_vm < VM_MAX) ? cur_vm : VM_OTHER]++;

        // only collect this while Windows running!!
        stats->iopl_count[iopl()]++;
    }

    if (intno == 0x21)      
        stats->int21[ah]++;
    else if (intno == 0x2f) 
    {
        stats->int2f[ah]++;
        if (ah == 0x16)
        {
            stats->int2f16[al]++;
            if (al == 7)    // 2f/1607/18 is most popular (VMPOLL)
                stats->int2f1607[(bx < 0x100) ? bx : VXD_OTHER]++;
        }
    }
}

void interrupt far int21(REG_PARAMS r)
{
    check_state(0x21, r.ax >> 8, r.ax & 0xff, r.bx);
    _chain_intr(old);
}

void interrupt far int2f(REG_PARAMS r)
{
    switch (r.ax)
    {
        case WIN_INIT_NOTIFY:   state = WIN_INIT_BEGIN; 
                                break;
        case WIN_INIT_COMPLETE: state = WIN_INIT_DONE; 
                                time(&stats->start); 
                                break;
        case WIN_EXIT_BEGIN:    state = WIN_FINI_BEGIN; 
                                time(&stats->end); 
                                break;
        case WIN_EXIT_NOTIFY:   state = WIN_FINI_DONE; 
                                break;
        case GET_STATS:         r.es = FP_SEG(fpstats); 
                                r.bx = FP_OFF(fpstats); 
                                break;
    }
	if (not_filter || (r.ax != 0x1607 && r.bx != 0x18))
        check_state(0x2f, r.ax >> 8, r.ax & 0xff, r.bx);
#if 1
	if (r.ax == 0x1607 && r.bx == 0x10 && r.cx == 3) {
		r.cx = 0;
		return;		// supports FastDisk
	}
#endif
    _chain_intr(old_2F);
}
