/*

VXDLIB.C	Windows95 W4 file compressor
Clive Turvey, Electronics Engineer, March 1995
Copyright (C) Tenth Planet Software Intl., Clive Turvey 1995.
All rights reserved. Non-Commercial use only.

Compiled using Microsoft C/C++ v8.00 (Visual C/C++ v1.00)

        cl -Ox vxdlib.c

******************************************************************************

 ====     ===
  \\      / th
   \\    /
    \\  /  Planet
     \\/    Software
     /\\     International
    /  \\
   /    \\
  /      \\
 ===     ====

     Copyright (C) Tenth Planet Software Intl., C Turvey 1995.
                       Chicago & Southampton.

                     CompuServe ID 74011,1732
                 Email : 74011.1732@compuserve.com

******************************************************************************

 Certain portions of this code and its algorithms remain the intellectual
 property of Clive Turvey, who retains the right to use and/or modify them
 at his discretion in projects for other clients.

******************************************************************************

Andrew Schulman sent me a hex dump of the W4 header, here is my description
of that W4 header, 0x12000 is arrived at from the new executable pointer in
the DOS EXE (The loader of the W4 file).

Offset  Content
0x12000   ASCII W       (Checked by loader)
0x12001   ASCII 4
0x12002   WORD 0x0004   (Windows version)
0x12004   WORD 0x2000   (Chunk size of 8K, Checked by loader, must be 8K)
0x12006   WORD 0x00E9   (Number of Chunks, Checked by loader must be less than 1K)
0x12008   ASCII D       (Compression Type? DoubleSpace signature?)
0x12009   ASCII S
0x1200A   0x00          ?
0x1200B   0x00          ?
0x1200C   0x00          ?
0x1200D   0x00          ?
0x1200E   0x00          ?
0x1200F   0x00          ?
0x12010   DWORD 0x123B4 File Offset to Start of Chunk 0x000
0x12014  .. 0xE7        Chunk Entries ..
0x123B0   DWORD 0xB49C2 File Offset to Start of Chunk 0x0E9
0x123B4   0x5E ....     Compressed Data for Chunk 0x000
0xB49C2 ....            Compressed Data for Chunk 0x0E9

The data is compressed using a Lempel Ziv algorithm with data encoded as
raw bytes or as a depth & count into a sliding history buffer. The following
table was derived from the Microsoft loader code in the DOS portion of
VMM32.VXD. It uses the same algorithm as DoubleSpace/DriveSpace. Software
to create W3 files and convert them between W3 and W4 is provided by Microsoft
in the Win95 DDK, the program is called DEVLIB. If you don't have Win95 you
could examine DBLSPACE.BIN or DRVSPACE.BIN. The Microsoft algorithm is
complicated because they have unfolded all the bit boundary loop, so it has
about eight times as much code to plough through. I have made no attempt to
optimize my algorithm, it would be a pointless exercise, and would not serve
my attempt to document it clearly. All the compressed chunks are stored in
a contiguous manner, with the last chunk being delimited by the end of the
file. It should be noted that DEVLIB strips the extranious resource junk
from the end of the VxD (LE) files it adds to the W3 library.


Shannon-Fano Tables for W4

Data encoded as Raw or Depth followed by Count

MSB...............LSB

            xxxxxxx01    Raw 1xxxxxxx, No Depth or Count
            xxxxxxx10    Raw 0xxxxxxx, No Depth or Count

                        Depth

             00000000    Quit
             xxxxxx00          x   (0..63)
             11111100    63
          00000000011    64
          xxxxxxxx011    64  + x  (64..319)
          11111111011    319
      000000000000111    320
      xxxxxxxxxxxx111    320 + x (320..4415)
      111111111111111    Check Buffer (Every 512 bytes)

                        Count

                    1    2
                  010    3
                  110    4
                00100    5
                01100    6
                10100    7
                11100    8
              0001000    9
              0011000    10
              0101000    11
              0111000    12
              1001000    13
              1011000    14
              1101000    15
              1111000    16
            000010000    17
            xxxx10000    17  + x  (17..32)
            111110000    32
          00000100000    33
          xxxxx100000    33  + x  (33..64)
          11111100000    64
        0000001000000    65
        xxxxxx1000000    65  + x  (65..128)
        1111111000000    128
      000000010000000    129
      xxxxxxx10000000    129 + x (129..256)
      111111110000000    256
    00000000100000000    257
    xxxxxxxx100000000    257 + x (257..512)
    11111111100000000    512
            000000000    Quit


*/

#include	<stdio.h>
#include	<string.h>
#include	<malloc.h>

typedef unsigned short WORD;
typedef unsigned char BYTE;
typedef unsigned long DWORD;

//#define DEBUG


// For an 8K History buffer HASHSHIFT = 5
//
// 2 ** (8+5) = 2 ** 13 = 8192
//
// The actual scope of the algorithm is 4414, a tad over 4K, but for a
// number of clever tricks to work, we need the next power of 2


#define HASHSHIFT 5

#define HASHSIZE  (1 << (8 + HASHSHIFT))

#define HASHMASK  (HASHSIZE - 1)

// One byte outside scope of algorithm

#define HASHMAX   (0xFFFF - HASHSIZE)

WORD hash_table[HASHSIZE];
WORD delta_list[HASHSIZE];


WORD compress(BYTE *, BYTE *, WORD);
WORD decompress(BYTE *, BYTE *, WORD);

WORD main(WORD,BYTE **);
void exit(int);
void getch(void);

#define EMAGIC					0x5A4D					// Old magic number
#define DSMAGIC					0x5344
#define W3MAGIC					0x3357
#define W4MAGIC					0x3457

struct exe_hdr													// DOS 1, 2, 3 .EXE header
{
	WORD	e_magic;				// Magic number
	WORD	e_cblp;					// Bytes on last page of file
	WORD	e_cp;						// Pages in file
	WORD	e_crlc;					// Relocations
	WORD	e_cparhdr;			// Size of header in paragraphs
	WORD	e_minalloc;			// Minimum extra paragraphs needed
	WORD	e_maxalloc;			// Maximum extra paragraphs needed
	WORD	e_ss;						// Initial (relative) SS value
	WORD	e_sp;						// Initial SP value
	WORD	e_csum;					// Checksum
	WORD	e_ip;						// Initial IP value
	WORD	e_cs;						// Initial (relative) CS value
	WORD	e_lfarlc;				// File address of relocation table
	WORD	e_ovno;					// Overlay number
	WORD	e_res[4];				// Reserved words
	WORD	e_oemid;				// OEM identifier (for e_oeminfo)
	WORD	e_oeminfo;			// OEM information; e_oemid specific
	WORD	e_res2[10];			// Reserved words
	DWORD	e_lfanew;				// File address of new exe header
} exeheader;

struct e32_exe													// New 32-bit .EXE header
{
	WORD	e32_magic;			// Magic number E32_MAGIC
	BYTE	e32_border;			// The byte ordering for the .EXE
	BYTE	e32_worder;			// The word ordering for the .EXE
	DWORD	e32_level;			// The EXE format level for now = 0
	WORD	e32_cpu;				// The CPU type
	WORD	e32_os;					// The OS type
	DWORD	e32_ver;				// Module version
	DWORD	e32_mflags;			// Module flags
	DWORD	e32_mpages;			// Module # pages
	DWORD	e32_startobj;		// Object # for instruction pointer
	DWORD	e32_eip;				// Extended instruction pointer
	DWORD	e32_stackobj;		// Object # for stack pointer
	DWORD	e32_esp;				// Extended stack pointer
	DWORD	e32_pagesize;		// .EXE page size
	DWORD	e32_pageshift;	// Page alignment shift in .EXE
	DWORD	e32_fixupsize;	// Fixup section size
	DWORD	e32_fixupsum;		// Fixup section checksum
	DWORD	e32_ldrsize;		// Loader section size
	DWORD	e32_ldrsum;			// Loader section checksum
	DWORD	e32_objtab;			// Object table offset
	DWORD	e32_objcnt;			// Number of objects in module
	DWORD	e32_objmap;			// Object page map offset
	DWORD	e32_itermap;		// Object iterated data map offset
	DWORD	e32_rsrctab;		// Offset of Resource Table
	DWORD	e32_rsrccnt;		// Number of resource entries
	DWORD	e32_restab;			// Offset of resident name table
	DWORD	e32_enttab;			// Offset of Entry Table
	DWORD	e32_dirtab;			// Offset of Module Directive Table
	DWORD	e32_dircnt;			// Number of module directives
	DWORD	e32_fpagetab;		// Offset of Fixup Page Table
	DWORD	e32_frectab;		// Offset of Fixup Record Table
	DWORD	e32_impmod;			// Offset of Import Module Name Table
	DWORD	e32_impmodcnt;	// Number of entries in Import Module Name Table
	DWORD	e32_impproc;		// Offset of Import Procedure Name Table
	DWORD	e32_pagesum;		// Offset of Per-Page Checksum Table
	DWORD	e32_datapage;		// Offset of Enumerated Data Pages
	DWORD	e32_preload;		// Number of preload pages
	DWORD	e32_nrestab;		// Offset of Non-resident Names Table
	DWORD	e32_cbnrestab;	// Size of Non-resident Name Table
	DWORD	e32_nressum;		// Non-resident Name Table Checksum
	DWORD	e32_autodata;		// Object # for automatic data object
	DWORD	e32_debuginfo;	// Offset of the debugging information
	DWORD	e32_debuglen;		// The length of the debugging info. in bytes
	DWORD	e32_instpreload;// Number of instance pages in preload section of .EXE file
	DWORD	e32_instdemand;	// Number of instance pages in demand load section of .EXE file
	DWORD	e32_heapsize;		// Size of heap - for 16-bit apps
	DWORD	e32_stacksize;	// Size of stack
	BYTE	e32_res3[20];
   // Pad structure to 196 bytes
} leheader;

struct w3_hdr
{
	WORD	w3_magic;
	WORD	w3_version;
	WORD	w3_entry_count;
	WORD	w4_unknown[5];
} w3header;

struct w3_dir
{
	BYTE	w3_entry_name[8];
	DWORD	w3_entry_offset;
	DWORD	w3_entry_le_hdr;
};

struct w4_hdr
{
	WORD	w4_magic;
	WORD	w4_version;
	WORD	w4_chunk_size;
	WORD	w4_chunk_count;
	WORD	w4_ds;
	WORD	w4_unknown[3];
} w4header;

BYTE exestub[0x80] = {
	0x4D,0x5A,0xFD,0x00,0x13,0x00,0x00,0x00,0x04,0x00,0x00,0x00,0xFF,0xFF,0x00,0x00,
	0xB8,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x40,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
	0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,
	0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x80,0x00,0x00,0x00,
	0x0E,0x1F,0xBA,0x0E,0x00,0xB4,0x09,0xCD,0x21,0xB8,0x01,0x4C,0xCD,0x21,0x54,0x68,
	0x69,0x73,0x20,0x70,0x72,0x6F,0x67,0x72,0x61,0x6D,0x20,0x63,0x61,0x6E,0x6E,0x6F,
	0x74,0x20,0x62,0x65,0x20,0x72,0x75,0x6E,0x20,0x69,0x6E,0x20,0x44,0x4F,0x53,0x20,
	0x6D,0x6F,0x64,0x65,0x2E,0x0D,0x0A,0x24,0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00 };

void pchar(byte)
BYTE byte;
{
	if ((byte > 31) && (byte < 127))
		putchar(byte);
	else
		putchar('.');
}

void dumpdata(len,buf,off)
DWORD len;
BYTE *buf;
DWORD off;
{
	DWORD i;

	while(len)
	{
		printf(" %08lX: ",off);

		for(i=0; i<16; i++)
		{
			if (i == 8)
				printf("- ");

			if (i < len)
				printf("%02X ",buf[i]);
			else
				printf("   ");
		}

		putchar(' ');

		for(i=0; i<16; i++)
			if (i < len)
				pchar(buf[i]);

		putchar('\n');

		if (len > 16)
			len -= 16;
		else
			len = 0;

		buf += 16;
		off += 16;
	}
}

// The following is a functional implementation of Microsoft's DoubleSpace
// decompression algorithm used for W4 multiple VXD files.

WORD decompress(source,dest,size)
BYTE *source;
BYTE *dest;
WORD size;
{
	WORD i, j, k, bitsused, bitsleft, count, depth;
	DWORD shiftreg;

	i = 0;

	shiftreg = 0;

	for(j=0; j<4; j++)
		shiftreg = (shiftreg >> 8l) + ((DWORD)*source++ << 24l);

	bitsleft = 8;

	depth = 1;

	while(depth)
	{
#ifdef DEBUG
		printf("%04X %02X:%08lX ",i,bitsleft,shiftreg);
#endif

		if (((shiftreg & 0x0003l) == 0x0001l) ||	// xxxxxxx01, Raw 1xxxxxxx, No Depth or Count
				((shiftreg & 0x0003l) == 0x0002l))		// xxxxxxx10, Raw 0xxxxxxx, No Depth or Count
		{
			if (--size == 0xFFFF) // Violation of Buffer
				return(1);

			dest[i] = (BYTE)(((shiftreg & 0x01FCl) >> 2l) | ((shiftreg & 0x0001l) << 7l));

#ifdef DEBUG
			printf("%02X ",dest[i]);

			pchar(dest[i]);

			printf("       ");
#endif

			i++;

			bitsused = 9;
		}
		else // Cooked
		{
			while(1) // Depth
			{
				if ((shiftreg & 0x0003l) == 0x0000l)	// xxxxxx00, x (0..63)
				{
					depth = (WORD)((shiftreg & 0x00FCl) >> 2);
					bitsused = 8;
					break;
				}

				if ((shiftreg & 0x0007l) == 0x0003l)	// xxxxxxxx011, 64 + x (64..319)
				{
					depth = (WORD)((shiftreg & 0x07F8l) >> 3) + 0x0040;
					bitsused = 11;
					break;
				}

				if ((shiftreg & 0x0007l) == 0x0007l)	// xxxxxxxxxxxx111, 320 + x (320..4415)
				{
					depth = (WORD)((shiftreg & 0x7FF8l) >> 3l) + 0x0140;
					bitsused = 15;
					break;
				}

				puts("Illegal Data");
				return(1);
			}

			if ((depth != 0x0000) &&	// 00000000 Quit
					(depth != 0x113F))		// 111111111111111 Check Buffer (Every 512 bytes)
			{
#ifdef DEBUG
				printf("Depth %4d ",depth);
#endif

				while(bitsused--) // skip used bits
				{
#ifdef DEBUG
					printf("%d",(WORD)(shiftreg & 1));
#endif

					shiftreg >>= 1;
					if (--bitsleft == 0)
					{
						shiftreg += (DWORD)*source++ << 24l;
						bitsleft = 8;
					}
				}

#ifdef DEBUG
				printf("\n                 ");
#endif

				while(1) // Count
				{
					if ((shiftreg & 0x00001l) == 0x00001l)	// 1, 2
					{
						count = 2;
						bitsused = 1;
						break;
					}

					if ((shiftreg & 0x00003l) == 0x00002l)	// x10, 3 + x (3..4)
					{
						count = (WORD)((shiftreg & 0x00004l) >> 2l) + 3;
						bitsused = 3;
						break;
					}

					if ((shiftreg & 0x00007l) == 0x00004l)	// xx100, 5 + x (5..8)
					{
						count = (WORD)((shiftreg & 0x0018) >> 3l) + 5;
						bitsused = 5;
						break;
					}

					if ((shiftreg & 0x0000Fl) == 0x00008l)	// xxx1000, 9 + x (9..16)
					{
						count = (WORD)((shiftreg & 0x00070l) >> 4l) + 9;
						bitsused = 7;
						break;
					}

					if ((shiftreg & 0x0001Fl) == 0x00010l)	// xxxx10000, 17 + x (17..32)
					{
						count = (WORD)((shiftreg & 0x001E0l) >> 5l) + 17;
						bitsused = 9;
						break;
					}

					if ((shiftreg & 0x0003Fl) == 0x00020l)	// xxxxx100000, 33 + x (33..64)
					{
						count = (WORD)((shiftreg & 0x007C0l) >> 6l) + 33;
						bitsused = 11;
						break;
					}

					if ((shiftreg & 0x0007Fl) == 0x00040l)	// xxxxxx1000000, 65 + x (65..128)
					{
						count = (WORD)((shiftreg & 0x01F80l) >> 7l) + 65;
						bitsused = 13;
						break;
					}

					if ((shiftreg & 0x000FFl) == 0x00080l)	// xxxxxxx10000000, 129 + x (129..256)
					{
						count = (WORD)((shiftreg & 0x07F00l) >> 8l) + 129;
						bitsused = 15;
						break;
					}

					if ((shiftreg & 0x001FFl) == 0x00100l)	// xxxxxxxx100000000, 257 + x (257..512)
					{
						count = (WORD)((shiftreg & 0x1FE00l) >> 9l) + 257;
						bitsused = 17;
						break;
					}

					depth = 0; // 000000000, Quit Condition
					count = 0;
					bitsused = 9;
					break; // Ugly Data
				}

#ifdef DEBUG
				printf("Count %4d ",count);
#endif

				k = count;

				while(k--)
				{
					if (--size == 0xFFFF) // Violation of Buffer
						return(1);

					dest[i] = dest[i-depth];
					i++;
				}
			}
			else
			{
#ifdef DEBUG
				if (depth == 0x0000)
					printf("Quit       ");	// No More data to decompress
				else
					printf("Chk Buffer ");	// At 512 byte sector boundary
#endif

				if ((depth == 0x113F) && // Check Buffer
						(size == 0x0000))
					depth = 0; // Quit Condition
			}
		}

		while(bitsused--) // skip used bits
		{
#ifdef DEBUG
			printf("%d",(WORD)(shiftreg & 1));
#endif

			shiftreg >>= 1;
			if (--bitsleft == 0)
			{
				shiftreg += (DWORD)*source++ << 24l;
				bitsleft = 8;
			}
		}

#ifdef DEBUG
		putchar('\n');
#endif
	}

	return(i);
}

// The following is a functional implementation of Microsoft's DoubleSpace
// compression algorithm used for W4 multiple VXD files. It should be noted
// that this code will probably fail if forced to compress more than 56k,
// this app only uses 8K clusters/chunks. This limition will be due to really
// stale data in the delta_list[], which I'm not clearing up. The problem
// is that a stale piece of data might appear to be valid as pointers wrap.
// So a) you could do some major garbage collection every 50k or so OR
// b) do 1% of the garbage collection every 512 bytes. I'm not going to do
// this here because it's pointless. Additionally there are a couple of
// things I could optimize here, but I think we'd lose a whole lot of
// clarity, notwithstanding we'll get a better compression ratio than DEVLIB
// we just get there slower, this is after all just 'C'.

WORD compress(source,dest,size)
BYTE *source;
BYTE *dest;
WORD size;
{
	WORD	i,
				retrolength,
				countlength,
				count,
				maxcount,
				maxdif,
				hash_code,
				hash_org,
				hash_depth,
				hash_delta,
				bitcount,
				newbits;

	BYTE	*ptr,
				*maxptr,
				*sourcestart,
				*sourceptr,
				*sourceend,
				*destptr;

	DWORD	shiftreg,
				newdata;

	sourceend		= source + size;

	sourcestart	= source;

	sourceptr		= sourcestart;

	destptr			= dest;

	for(i=0; i<HASHSIZE; i++)	// Void hash table by pointing to data outside
	{													// the scope of the compression algorithm
		hash_table[i] = HASHMAX;
		delta_list[i] = HASHMAX;
	}

	bitcount = 0;	// Unflushed bits within the encoding register

	shiftreg = 0;	// Encoding register


	while(sourceptr < sourceend)
	{
		retrolength = (WORD)(sourceptr - sourcestart);

		if (retrolength > 4414) // Limit ourselves to what we can actually encode
			retrolength = 4414;

		countlength = 0x0200 - ((WORD)(sourceptr - sourcestart) & 0x01FF); // Sectorize

#ifdef DEBUG
		printf("%3ld %4ld (%02X)\t",countlength,retrolength,*sourceptr);
#endif

		ptr = sourceptr;

		maxptr   = ptr;

		maxcount = 1;

		if (countlength > 1)
		{
			hash_code = (sourceptr[0] ^ (sourceptr[1] << HASHSHIFT)); // Calculate Hashing code

			hash_depth = (WORD)(sourceptr - sourcestart);	// Determine current offset

			hash_org = hash_table[hash_code];		// Retrieve Previous Hash match

			hash_table[hash_code] = hash_depth;	// Replace with new match

			hash_delta = hash_depth - hash_org;	// Calculate delta between the two

			if (hash_delta > HASHSIZE)
				hash_delta = HASHMAX;	// data outside the scope of compression algorithm

			delta_list[(hash_depth & HASHMASK)] = hash_delta;	// And add to our chain of matches

			while(hash_delta <= retrolength)	// Beyond our scope?
			{
				ptr = sourceptr - hash_delta;		// Timewarp back in history

				for(count=0; count < countlength; count++)	// See how big a match
				{
					if (sourceptr[count] != ptr[count])	// Compare or Die
						break;
				}

				if (count > maxcount)	// Best match yet
				{
					maxcount = count;
					maxptr   = ptr;

					if (maxcount == countlength)	// Won't get any better than this
						break;
				}

				hash_depth = (WORD)(ptr - sourcestart);	// Determine current offset

				hash_delta = hash_delta + delta_list[(hash_depth & HASHMASK)];
			}

			maxdif = (WORD)(sourceptr - maxptr);
		}

//	putchar('\n');

// Below Optional, Adds Cooked bytes to Hash Table. We can choose to hash
// only the initial two bytes, doing so will reduce the amount of data in
// the hash table and speed up compression. It will however mean that some
// matchs that are in the data stream won't be in the hash table and
// compression ratios will drop.

		if (maxcount != 1)
		{ // watch out for last byte senerio
			for(count = 1; count < maxcount; count++)
			{
				hash_code = (sourceptr[count+0] ^ (sourceptr[count+1] << HASHSHIFT)); // Calculate Hashing code

				hash_depth = (WORD)(sourceptr - sourcestart) + count;	// Determine current offset

				hash_org = hash_table[hash_code];		// Retrieve Previous Hash match

				hash_table[hash_code] = hash_depth;	// Replace with new match

				hash_delta = hash_depth - hash_org;	// Calculate delta between the two

				if (hash_delta > HASHSIZE)
					hash_delta = HASHMAX; // data outside the scope of compression algorithm

				delta_list[(hash_depth & HASHMASK)] = hash_delta;	// And add to our chain of matches
			}
		}

// Above Optional

		// Encoding Section

// In the situation where the maxcount is two, any depth beyond 63 is more
// efficiently encoded as two raw bytes.

		if ((maxcount == 1) ||
				((maxcount == 2) && (maxdif > 63)))
		{
			// Raw Encoding

#ifdef DEBUG
			printf("Raw\t\t");
#endif

			for (i=0; i<maxcount; i++)
			{
				newbits = 9;

				newdata = (((DWORD)sourceptr[i] & 0x7Fl) << 2l);

				if (sourceptr[i] & 0x80)
					newdata |= 0x01l;
				else
					newdata |= 0x02l;

				shiftreg |= newdata << bitcount;
				bitcount += newbits;
			}
		}
		else
		{
			// Cooked Encoding

#ifdef DEBUG
			printf("Cooked %4ld,%3ld\t",maxdif,maxcount);
#endif

// Encode Depth First

			if ((maxdif >= 1) && (maxdif <= 63))	// xxxxxx00 : x (1..63)
			{
				newbits = 8;
				newdata = ((DWORD)maxdif << 2l);
			}

			if ((maxdif >= 64) && (maxdif <= 319))	// xxxxxxxx011 : 64 + x (64..319)
			{
				newbits = 11;
				newdata = (((DWORD)maxdif - 64l) << 3l) | 0x03l;
			}

			if ((maxdif >= 320) && (maxdif <= 4414))	// xxxxxxxxxxxx111 : 320 + x (320..4414)
			{
				newbits = 15;
				newdata = (((DWORD)maxdif - 320l) << 3l) | 0x07l;
			}

	if (maxdif > 4414)
		exit(1);

			shiftreg |= newdata << bitcount;
			bitcount += newbits;

#ifdef DEBUG
	i = newbits;
#endif

			while(bitcount > 8) // Flush whole bytes from the encoding register
			{
				*destptr++ = (BYTE)(shiftreg & 0x000000FFl);
				shiftreg >>= 8;
				bitcount -= 8;
			}

// Encode Count Second, a lookup table would work well here

			if (maxcount == 2)	// 1 : 2
			{
				newbits = 1;
				newdata = 0x01l;
			}

			if ((maxcount >= 3) && (maxcount <= 4))	// x10 : 3 + x (3..4)
			{
				newbits = 3;
				newdata = (((DWORD)maxcount - 3l) << 2l) | 0x02l;
			}

			if ((maxcount >= 5) && (maxcount <= 8))	// xx100 : 5 + x (5..8)
			{
				newbits = 5;
				newdata = (((DWORD)maxcount - 5l) << 3l) | 0x04l;
			}

			if ((maxcount >= 9) && (maxcount <= 16))	// xxx1000 : 9 + x (9..16)
			{
				newbits = 7;
				newdata = (((DWORD)maxcount - 9l) << 4l) | 0x08l;
			}

			if ((maxcount >= 17) && (maxcount <= 32))	// xxxx10000 : 17 + x (17..32)
			{
				newbits = 9;
				newdata = (((DWORD)maxcount - 17l) << 5l) | 0x10l;
			}

			if ((maxcount >= 33) && (maxcount <= 64))	// xxxxx100000 : 33 + x (33..64)
			{
				newbits = 11;
				newdata = (((DWORD)maxcount - 33l) << 6l) | 0x20l;
			}

			if ((maxcount >= 65) && (maxcount <= 128))	// xxxxxx1000000 : 65 + x (65..128)
			{
				newbits = 13;
				newdata = (((DWORD)maxcount - 65l) << 7l) | 0x40l;
			}

			if ((maxcount >= 129) && (maxcount <= 256))	// xxxxxxx10000000 : 129 + x (129..256)
			{
				newbits = 15;
				newdata = (((DWORD)maxcount - 129l) << 8l) | 0x80l;
			}

			if ((maxcount >= 257) && (maxcount <= 512))	// xxxxxxxx100000000 : 257 + x (257..512)
			{
				newbits = 17;
				newdata = (((DWORD)maxcount - 257l) << 9l) | 0x100l;
			}

			shiftreg |= newdata << bitcount;
			bitcount += newbits;

#ifdef DEBUG
			i += newbits;

			printf("%d ",i);
#endif

		}

		while(bitcount > 8) // Flush whole bytes from the encoding register
		{
			*destptr++ = (BYTE)(shiftreg & 0x000000FFl);
			shiftreg >>= 8;
			bitcount -= 8;
		}

#ifdef DEBUG
		for(i=0; i<maxcount; i++)
		{
			if (i == 0)
				printf("(%02X",sourceptr[i]);
			else
				printf(" %02X",sourceptr[i]);
		}
		puts(")");
#endif

		sourceptr += maxcount;

		if (((WORD)(sourceptr - sourcestart) & 0x01FF) == 0) // Check sector boundary
		{
#ifdef DEBUG
			puts("\t\tCheck Buffer");
#endif

			newbits = 15;				// Encode Check Buffer
			newdata = 0x7FFFl;
			shiftreg |= newdata << bitcount;
			bitcount += newbits;

			while(bitcount > 8)	// Flush whole bytes from the encoding register
			{
				*destptr++ = (BYTE)(shiftreg & 0x000000FFl);
				shiftreg >>= 8;
				bitcount -= 8;
			}
		}
	}

#ifdef DEBUG
	puts("\t\tQuit");
#endif

	newbits = 8;				// Encode Quit
	newdata = 0x00l;
	shiftreg |= newdata << bitcount;
	bitcount += newbits;

	while(bitcount > 8)	// Flush whole bytes from the encoding register
	{
		*destptr++ = (BYTE)(shiftreg & 0x000000FFl);
		shiftreg >>= 8;
		bitcount -= 8;
	}

	*destptr++ = (BYTE)(shiftreg & 0x000000FFl); // Flush final bits

	return((WORD)(destptr - dest));
}


#define CHK_CANTOPEN	0x0001
#define CHK_NOTEXE		0x0002
#define CHK_NOTNEWEXE	0x0003
#define CHK_NOTWEXE		0x0004
#define CHK_W3				0x0010
#define CHK_W4				0x0020

WORD chk_file(filename)
BYTE *filename;
{
	FILE *f;
	DWORD newexeoffset;

	f = fopen(filename,"rb");

	if (!f)
		return(CHK_CANTOPEN);

	if (fread(&exeheader,2,sizeof(exeheader),f) != sizeof(exeheader))
	{
		fclose(f);
		return(CHK_NOTEXE);
	}

	if (exeheader.e_magic != EMAGIC)
	{
		fclose(f);
		return(CHK_NOTEXE);
	}

	if ((exeheader.e_lfarlc >= 0x40) ||
			((exeheader.e_lfarlc + (4 * exeheader.e_crlc)) < 0x3C))
	{
		newexeoffset = exeheader.e_lfanew;
	}
	else
	{
		fclose(f);
		return(CHK_NOTNEWEXE);
	}

	fseek(f,newexeoffset,0);

	/* Load W3 header, 0x10 bytes */

	if (fread(&w3header,1,sizeof(w3header),f) != sizeof(w3header))
	{
		fclose(f);
		return(CHK_NOTNEWEXE);
	}

	fclose(f);

	if (w3header.w3_magic == W3MAGIC) /* Check W3 signature */
		return(CHK_W3);

	if (w3header.w3_magic == W4MAGIC) /* Check W4 signature */
		return(CHK_W4);

	return(CHK_NOTWEXE);
}


WORD dump_vxd(w3file)
BYTE *w3file;
{
	FILE *fwthree;
	WORD i, j;
	struct w3_dir *w3directory;
	DWORD newexeoffset;

	fwthree = fopen(w3file,"rb");

	if (!fwthree)
	{
		printf("Can't open file '%s'\n",w3file);
		return(1);
	}

	if (fread(&exeheader,2,sizeof(exeheader),fwthree) != sizeof(exeheader))
	{
		printf("Unable to read .EXE header from '%s'.\n",w3file);
		fclose(fwthree);
		return(1);
	}

	if (exeheader.e_magic != EMAGIC)
	{
		printf("'%s' not a .EXE file.\n",w3file);
		return(1);
	}

	if ((exeheader.e_lfarlc >= 0x40) ||
			((exeheader.e_lfarlc + (4 * exeheader.e_crlc)) < 0x3C))
		newexeoffset = exeheader.e_lfanew;
	else
	{
		printf("'%s' does not appear to be a Windows W3/W4 file.\n",w3file);
		return(1);
	}

	fseek(fwthree,newexeoffset,0);

	/* Load W3 header, 0x10 bytes */

	if (fread(&w3header,1,sizeof(w3header),fwthree) != sizeof(w3header))
	{
		printf("Unable to read W3 header from '%s'.\n",w3file);
		fclose(fwthree);
		return(1);
	}

	if (w3header.w3_magic == W4MAGIC) /* Check W4 signature */
	{
		printf("'%s' appears to be a Windows W4 file.\n",w3file);
		return(1);
	}

	if (w3header.w3_magic != W3MAGIC) /* Check W3 signature */
	{
		printf("'%s' does not appear to be a Windows W3 file.\n",w3file);
		return(1);
	}

	w3directory = malloc(w3header.w3_entry_count * sizeof(struct w3_dir));

	if(w3directory == NULL)
	{
		printf("Insufficent directory memory. (%d Bytes required)\n",(w3header.w3_entry_count * sizeof(struct w3_dir)));
		return(1);
	}

	if (fread(w3directory,sizeof(struct w3_dir),w3header.w3_entry_count,fwthree) != w3header.w3_entry_count)
	{
		printf("Unable to read W3 directory from '%s'.\n",w3file);
		fclose(fwthree);
		return(1);
	}

	printf("Dumping %s\n\nvxds:\n",w3file);

	for(i=0; i<w3header.w3_entry_count; i++)
	{
		printf("  ");

		for(j=0; j<8; j++)
			putchar(w3directory[i].w3_entry_name[j]);

		printf(" : LE header offset %8lX, size %8lX\n",
			w3directory[i].w3_entry_offset,
			w3directory[i].w3_entry_le_hdr);
	}

	printf("%3d total vxd(s)\n",w3header.w3_entry_count);

	free(w3directory);

	fclose(fwthree);

	return(0);
}


WORD extract_vxd(w3file,vmmname)
BYTE *w3file, *vmmname;
{
	FILE *fwthree, *fvmm;
	WORD i, j, k;
	struct w3_dir *w3directory;
	DWORD newexeoffset, start, finish, size;

	fwthree = fopen(w3file,"rb");

	if (!fwthree)
	{
		printf("Can't open file '%s'\n",w3file);
		return(1);
	}

	if (fread(&exeheader,2,sizeof(exeheader),fwthree) != sizeof(exeheader))
	{
		printf("Unable to read .EXE header from '%s'.\n",w3file);
		fclose(fwthree);
		return(1);
	}

	if (exeheader.e_magic != EMAGIC)
	{
		printf("'%s' not a .EXE file.\n",w3file);
		return(1);
	}

	if ((exeheader.e_lfarlc >= 0x40) ||
			((exeheader.e_lfarlc + (4 * exeheader.e_crlc)) < 0x3C))
		newexeoffset = exeheader.e_lfanew;
	else
	{
		printf("'%s' does not appear to be a Windows W3/W4 file.\n",w3file);
		return(1);
	}

	fseek(fwthree,newexeoffset,0);

	/* Load W3 header, 0x10 bytes */

	if (fread(&w3header,1,sizeof(w3header),fwthree) != sizeof(w3header))
	{
		printf("Unable to read W3 header from '%s'.\n",w3file);
		fclose(fwthree);
		return(1);
	}

	if (w3header.w3_magic == W4MAGIC) /* Check W4 signature */
	{
		printf("'%s' appears to be a Windows W4 file.\n",w3file);
		return(1);
	}

	if (w3header.w3_magic != W3MAGIC) /* Check W3 signature */
	{
		printf("'%s' does not appear to be a Windows W3 file.\n",w3file);
		return(1);
	}

	w3directory = malloc(w3header.w3_entry_count * sizeof(struct w3_dir));

	if(w3directory == NULL)
	{
		printf("Insufficent directory memory. (%d Bytes required)\n",(w3header.w3_entry_count * sizeof(struct w3_dir)));
		return(1);
	}

	if (fread(w3directory,sizeof(struct w3_dir),w3header.w3_entry_count,fwthree) != w3header.w3_entry_count)
	{
		printf("Unable to read W3 directory from '%s'.\n",w3file);
		fclose(fwthree);
		return(1);
	}

	printf("Extracting %s to %s\n",w3file,vmmname);

	for(i=0; i<w3header.w3_entry_count; i++)
	{
		for(j=0; j<8; j++)
		{

			if ((w3directory[i].w3_entry_name[j] == 0x20) &&
					((vmmname[j] == 0x00) || (vmmname[j] == '.')))
			{
				j = 8;
				break;
			}

			if ((w3directory[i].w3_entry_name[j] != vmmname[j]) &&
					(w3directory[i].w3_entry_name[j] != (vmmname[j] & 0xDF)))
				break;
		}

		if (j == 8)
		{
			start  = w3directory[i].w3_entry_offset;

			if ((i+1) == w3header.w3_entry_count)
			{
				fseek(fwthree,0l,SEEK_END);

				finish = ftell(fwthree);
			}
			else
			{
				finish = w3directory[i+1].w3_entry_offset;
			}

			size = (finish - start);

			printf("%s %08lX %08lX %08lX\n",vmmname,start,finish,size);

			fseek(fwthree,start,SEEK_SET);

			fvmm = fopen(vmmname,"wb");

			if (!fvmm)
			{
				printf("Can't open file '%s'\n",vmmname);
				return(1);
			}

			k = (WORD)((size + 0x80l) & 0x01FFl);

			exestub[2] = (BYTE)(k & 0x00FF);
			exestub[3] = (BYTE)(k >> 8);

			k = (WORD)((size + 0x80l + 0x01FFl) >> 9l);

			exestub[4] = (BYTE)(k & 0x00FF);
			exestub[5] = (BYTE)(k >> 8);

			for(k=0; k<0x80; k++)
				fputc(exestub[k],fvmm);

			if (fread(&leheader,1,sizeof(leheader),fwthree) != sizeof(leheader))
			{
				printf("Unable to read LE header inside '%s'.\n",w3file);
				fclose(fwthree);
				return(1);
			}

			leheader.e32_datapage -= (start - 0x80l);

			if (fwrite(&leheader,1,sizeof(leheader),fvmm) != sizeof(leheader))
			{
				printf("Unable to write LE header to '%s'.\n",vmmname);
				fclose(fvmm);
				return(1);
			}

			size -= sizeof(leheader);

			while(size--)
				fputc(fgetc(fwthree),fvmm);

			fclose(fvmm);

			break;
		}
	}

	if (i == w3header.w3_entry_count)
		puts("VxD not found.");

	free(w3directory);

	fclose(fwthree);

	return(0);
}


WORD pack_vxd(w3file,w4file)
BYTE *w3file, *w4file;
{
	FILE *fwthree, *fwfour;
	BYTE *source, *dest, *ptr;
	WORD newsize, size;
	DWORD i, j, k, l, start, finish;
	DWORD totalnewsize, newexeoffset, exesize;
	DWORD *chunk_table;

	fwthree = fopen(w3file,"rb");

	if (!fwthree)
	{
		printf("Can't open file '%s'\n",w3file);
		return(1);
	}

	if (fread(&exeheader,2,sizeof(exeheader),fwthree) != sizeof(exeheader))
	{
		printf("Unable to read .EXE header from '%s'.\n",w3file);
		fclose(fwthree);
		return(1);
	}

	if (exeheader.e_magic != EMAGIC)
	{
		printf("'%s' not a .EXE file.\n",w3file);
		return(1);
	}

	if ((exeheader.e_lfarlc >= 0x40) ||
			((exeheader.e_lfarlc + (4 * exeheader.e_crlc)) < 0x3C))
		newexeoffset = exeheader.e_lfanew;
	else
	{
		printf("'%s' does not appear to be a Windows W3/W4 file.\n",w3file);
		return(1);
	}

	fseek(fwthree,newexeoffset,0);

	/* Load W3 header, 0x10 bytes */

	if (fread(&w3header,1,sizeof(w3header),fwthree) != sizeof(w3header))
	{
		printf("Unable to read W3 header from '%s'.\n",w3file);
		fclose(fwthree);
		return(1);
	}

	if (w3header.w3_magic == W4MAGIC) /* Check W4 signature */
	{
		printf("'%s' appears to be a Windows W4 file.\n",w3file);
		return(1);
	}

	if (w3header.w3_magic != W3MAGIC) /* Check W3 signature */
	{
		printf("'%s' does not appear to be a Windows W3 file.\n",w3file);
		return(1);
	}

	fwfour = fopen(w4file,"wb");

	if (!fwfour)
	{
		printf("Can't open file '%s'.\n",w4file);
		return(1);
	}

	fseek(fwthree,0l,SEEK_END);

	exesize = ftell(fwthree) - newexeoffset;

	printf("Packing %s to %s\n\n",w3file,w4file);

	fseek(fwthree,0l,SEEK_SET);

	i = newexeoffset;

	while(i--)
		fputc(fgetc(fwthree),fwfour);

	w4header.w4_magic				= W4MAGIC;

	w4header.w4_version			= w3header.w3_version;

	w4header.w4_ds					= DSMAGIC;

	w4header.w4_chunk_size	= 0x2000;

	w4header.w4_chunk_count	= (WORD)((exesize + w4header.w4_chunk_size - 1) /
															w4header.w4_chunk_size);

	w4header.w4_unknown[0]	= 0;
	w4header.w4_unknown[1]	= 0;
	w4header.w4_unknown[2]	= 0;

// Move file pointer to account for W4 header and chunk table

	fseek(fwfour,(newexeoffset + 0x10l + ((DWORD)w4header.w4_chunk_count * 4l)),SEEK_SET);

	if (w4header.w4_chunk_size > (WORD)(HASHMAX))
	{
		printf("The file is too big, Compiled limit %u bytes.\n",HASHMAX);
		return(1);
	}

	source = malloc(w4header.w4_chunk_size);

	if(source == NULL)
	{
		printf("Insufficent source buffer memory. (%d Bytes required)\n",w4header.w4_chunk_size);
		return(1);
	}

	dest = malloc(w4header.w4_chunk_size);

	if(dest == NULL)
	{
		printf("Insufficent destination buffer memory. (%d Bytes required)\n",w4header.w4_chunk_size);
		return(1);
	}

	chunk_table = malloc(w4header.w4_chunk_count * 4);

	if(chunk_table == NULL)
	{
		printf("Insufficent chunk memory. (%d Bytes required)\n",(w4header.w4_chunk_count * 4));
		return(1);
	}

	totalnewsize	= 0l;

	j = w4header.w4_chunk_count;

	for(k=0; k<j; k++)
	{
		start  = ftell(fwthree);

		if ((k+1) == w4header.w4_chunk_count)
		{
			i		= w4header.w4_chunk_size;

			ptr	= source;

			while(i--)
				*ptr++ = 0x00;

			fseek(fwthree,0l,SEEK_END);

			finish = ftell(fwthree);

			fseek(fwthree,start,SEEK_SET);
		}
		else
		{
			finish = start + (DWORD)w4header.w4_chunk_size;
		}

		size = (WORD)(finish - start);

		if (fread(source,1,size,fwthree) != size)
		{
			printf("Unable to read W3 file '%s'.\n",w3file);
			fclose(fwthree);
			return(1);
		}

		chunk_table[k] = ftell(fwfour);

		newsize = compress(source,dest,w4header.w4_chunk_size);

		if (newsize >= w4header.w4_chunk_size)
		{
			ptr = source;

			newsize = w4header.w4_chunk_size;
		}
		else
		{
			ptr = dest;
		}

		if (fwrite(ptr,1,newsize,fwfour) != newsize)
		{
			printf("Unable to write W4 file '%s'.\n",w4file);
			fclose(fwfour);
			return(1);
		}

		totalnewsize	+= newsize;

		fprintf(stderr,"  %3ld%% complete\r",(((k+1) * 100) / j));
	}

	i = exesize + newexeoffset;

	j = exesize;

	k = totalnewsize;

	l = k + newexeoffset + 0x10l + ((DWORD)w4header.w4_chunk_count * 4l);

	printf("%-40s%ld\n","Size of library:",i);

	printf("%-40s%ld\n","Size of compressible portion:",j);

	printf("%-40s%d\n","Size of chunk map:",(w4header.w4_chunk_count * 4));

	printf("%-40s%ld (%d%% of original)\n","New size of compressible portion:",k,((k * 100) / j));

	printf("%-40s%ld (%d%% of original)\n","New size of library:",l,((l * 100) / i));


	fseek(fwfour,newexeoffset,SEEK_SET);

// Write W4 header & chunk table

	if (fwrite(&w4header,1,sizeof(w4header),fwfour) != sizeof(w4header))
	{
		printf("Unable to write W4 header to '%s'.\n",w3file);
		fclose(fwfour);
		return(1);
	}

	if (fwrite(chunk_table,4,w4header.w4_chunk_count,fwfour) != w4header.w4_chunk_count)
	{
		printf("Unable to write W4 chunk table to '%s'.\n",w3file);
		fclose(fwfour);
		return(1);
	}

	free(source);

	free(dest);

	free(chunk_table);

	fclose(fwthree);

	fclose(fwfour);

	return(0);
}


WORD unpack_vxd(w4file,w3file)
BYTE *w4file, *w3file;
{
	FILE *fwfour, *fwthree;
	BYTE *source, *dest, *ptr;
	WORD size;
	DWORD i, j, k, l, start, finish;
	DWORD newexeoffset, exesize;
	DWORD *chunk_table;

	fwfour = fopen(w4file,"rb");

	if (!fwfour)
	{
		printf("Can't open file '%s'\n",w4file);
		return(1);
	}

	if (fread(&exeheader,2,sizeof(exeheader),fwfour) != sizeof(exeheader))
	{
		printf("Unable to read .EXE header from '%s'.\n",w4file);
		fclose(fwfour);
		return(1);
	}

	if (exeheader.e_magic != EMAGIC)
	{
		printf("'%s' not a .EXE file.\n",w4file);
		return(1);
	}

	if ((exeheader.e_lfarlc >= 0x40) ||
			((exeheader.e_lfarlc + (4 * exeheader.e_crlc)) < 0x3C))
		newexeoffset = exeheader.e_lfanew;
	else
	{
		printf("'%s' does not appear to be a Windows W3/W4 file.\n",w4file);
		return(1);
	}

	fseek(fwfour,newexeoffset,0);

	/* Load W4 header, 0x10 bytes */

	if (fread(&w4header,1,sizeof(w4header),fwfour) != sizeof(w4header))
	{
		printf("Unable to read W4 header from '%s'.\n",w4file);
		fclose(fwfour);
		return(1);
	}

	if (w4header.w4_magic == W3MAGIC) /* Check W3 signature */
	{
		printf("'%s' appears to be a Windows W3 file.\n",w4file);
		return(1);
	}

	if (w4header.w4_magic != W4MAGIC) /* Check W4 signature */
	{
		printf("'%s' does not appear to be a Windows W4 file.\n",w4file);
		return(1);
	}

	chunk_table = malloc(w4header.w4_chunk_count * 4);

	if(chunk_table == NULL)
	{
		printf("Insufficent chunk memory. (%d Bytes required)\n",(w4header.w4_chunk_count * 4));
		return(1);
	}

	if (fread(chunk_table,4,w4header.w4_chunk_count,fwfour) != w4header.w4_chunk_count)
	{
		printf("Unable to read W4 chunk table from '%s'.\n",w4file);
		fclose(fwfour);
		return(1);
	}

	fwthree = fopen(w3file,"wb");

	if (!fwthree)
	{
		printf("Can't open file '%s'.\n",w3file);
		return(1);
	}

	fseek(fwfour,0l,SEEK_END);

	exesize = ftell(fwfour) - newexeoffset;

	printf("Unpacking %s to %s\n\n",w4file,w3file);

	fseek(fwfour,0l,SEEK_SET);

	i = newexeoffset;

	while(i--)
		fputc(fgetc(fwfour),fwthree);

	if (w4header.w4_chunk_size > (WORD)(HASHMAX))
	{
		printf("The file is too big, Compiled limit %u bytes.\n",HASHMAX);
		return(1);
	}

	source = malloc(w4header.w4_chunk_size);

	if(source == NULL)
	{
		printf("Insufficent source buffer memory. (%d Bytes required)\n",w4header.w4_chunk_size);
		return(1);
	}

	dest = malloc(w4header.w4_chunk_size);

	if(dest == NULL)
	{
		printf("Insufficent destination buffer memory. (%d Bytes required)\n",w4header.w4_chunk_size);
		return(1);
	}

	j = w4header.w4_chunk_count;

	for(k=0; k<j; k++)
	{
		start  = chunk_table[k];

		if ((k+1) == w4header.w4_chunk_count)
		{
			fseek(fwfour,0l,SEEK_END);

			finish = ftell(fwfour);
		}
		else
		{
			finish = chunk_table[k+1];
		}

		size = (WORD)(finish - start);

		fseek(fwfour,start,SEEK_SET);

		if (fread(source,1,size,fwfour) != size)
		{
			printf("Unable to read W4 file '%s'.\n",w4file);
			fclose(fwfour);
			return(1);
		}

		if (size == w4header.w4_chunk_size)
		{
			ptr = source;
		}
		else
		{
			if (decompress(source,dest,w4header.w4_chunk_size) != w4header.w4_chunk_size)
			{
				puts("Decompression Error.");
				return(1);
			}

			ptr = dest;
		}

		if (fwrite(ptr,1,w4header.w4_chunk_size,fwthree) != w4header.w4_chunk_size)
		{
			printf("Unable to write W3 file '%s'.\n",w3file);
			fclose(fwthree);
			return(1);
		}

		fprintf(stderr,"  %3ld%% complete\r",(((k+1) * 100) / j));
	}

	free(source);

	free(dest);

	free(chunk_table);

	fclose(fwfour);

	fclose(fwthree);

	return(0);
}


WORD main(argc,argv)
WORD argc;
BYTE *argv[];
{
	WORD i, j, dump_option, pack_option, unpack_option, extract_option;
	BYTE tmp_name[] = "VxDLIB.$$$";

	fprintf(stderr,"\nVxDLIB v0.05 (c) Copyright Tenth Planet Software Intl., C Turvey 1995.\n"
						"\t\t\t   All rights reserved. Non-Commercial use only.\n\n");

	if (argc < 2)
	{
		puts(	" ====     ===\t\t\tThis FREEWARE product was written by Clive\n"
					"  \\\\      / th\t\t\tTurvey, an English Electronics Engineer.\n"
					"   \\\\    /");
		puts(	"    \\\\  /   Planet\t\tIt is designed to replace the DEVLIB utility\n"
					"     \\\\/     Software\t\tprovided by Microsoft in the Win95 DDK.\n"
					"     /\\\\      International");
		puts(	"    /  \\\\\t\t\tIf you like the software, or find a problem\n"
					"   /    \\\\   CIS 74011,1732\tpost some Email to 74011.1732@compuserve.com\n"
					"  /      \\\\\n"
					" ===     ====\n");

		puts("Usage : VxDLIB <-d[ump]> <W3/W4 file>\n"
					"\tVxDLIB <-p[ack]> <W3/W4 file>\n"
					"\tVxDLIB <-u[npack]> <W4 file>\n"
					"\tVxDLIB <-e[xtract]> <W3/W4 file> <VxD>\n");

		puts("!!!BETA!!! This software is a beta release !!!BETA!!!");

		exit(1);
	}

	dump_option = 0;

	pack_option = 0;

	unpack_option = 0;

	extract_option = 0;

	j = 1;

	for(i=1; i<argc; i++)
	{
		if ((*argv[i] == '/') || (*argv[i] == '-'))
		{
			if (strnicmp(argv[i]+1, "dump",1) == 0)
			{
				dump_option++;

				continue;
			}

			if (strnicmp(argv[i]+1, "pack",1) == 0)
			{
				pack_option++;

				continue;
			}

			if (strnicmp(argv[i]+1, "unpack",1) == 0)
			{
				unpack_option++;

				continue;
			}

			if (strnicmp(argv[i]+1, "extract",1) == 0)
			{
				extract_option++;

				continue;
			}

			printf("Illegal option : %s\n",argv[i]);
			exit(1);
		}
		else
			argv[j++] = argv[i];
	}

	argc = j;

	if ((dump_option + pack_option + unpack_option + extract_option) > 1)
	{
		puts("VxDLIB Error : Too many options, pick one");
		exit(1);
	}

	if (dump_option)
	{
		if (argc < 2)
		{
			puts("VxDLIB Error : DUMP : Insufficient parameters");
			exit(1);
		}

		switch(chk_file(argv[1]))
		{
			case CHK_CANTOPEN :
				puts("VxDLIB Error : DUMP : Can't open file");
				break;

			case CHK_NOTEXE :
				puts("VxDLIB Error : DUMP : File is not an EXEcutable");
				break;

			case CHK_NOTNEWEXE :
				puts("VxDLIB Error : DUMP : File is not a New EXEcutable");
				break;

			case CHK_NOTWEXE :
				puts("VxDLIB Error : DUMP : File is not a W? EXEcutable");
				break;

			case CHK_W3 :
				dump_vxd(argv[1]);
				break;

			case CHK_W4 :
				if (!unpack_vxd(argv[1],tmp_name))
				{
					dump_vxd(tmp_name);
					unlink(tmp_name);
				}

				break;

			default :
				puts("VxDLIB Error : DUMP : Unsuitable file");
		}
	}

	if (pack_option)
	{
		if (argc < 2)
		{
			puts("PACK : Insufficient parameters");
			exit(1);
		}

		switch(chk_file(argv[1]))
		{
			case CHK_CANTOPEN :
				puts("VxDLIB Error : PACK : Can't open file");
				break;

			case CHK_NOTEXE :
				puts("VxDLIB Error : PACK : File is not an EXEcutable");
				break;

			case CHK_NOTNEWEXE :
				puts("VxDLIB Error : PACK : File is not a New EXEcutable");
				break;

			case CHK_NOTWEXE :
				puts("VxDLIB Error : PACK : File is not a W? EXEcutable");
				break;

			case CHK_W3 :
				if (access(argv[1],0x06)) // Read/Write Permission
				{
					puts("VxDLIB Error : PACK : File does not have Read/Write permission");
				}
				else
				{
					if(!pack_vxd(argv[1],tmp_name))
					{
						unlink(argv[1]);
						rename(tmp_name,argv[1]);
					}
				}

				break;

			case CHK_W4 :
				if (access(argv[1],0x06)) // Read/Write Permission
				{
					puts("VxDLIB Error : PACK : File does not have Read/Write permission");
				}
				else
				{
					if (!unpack_vxd(argv[1],tmp_name))
					{
						pack_vxd(tmp_name,argv[1]);
						unlink(tmp_name);
					}
				}
				break;

			default :
				puts("VxDLIB Error : PACK : Unsuitable file");
		}
	}

	if (unpack_option)
	{
		if (argc < 2)
		{
			puts("UNPACK : Insufficient parameters");
			exit(1);
		}

		switch(chk_file(argv[1]))
		{
			case CHK_CANTOPEN :
				puts("VxDLIB Error : UNPACK : Can't open file");
				break;

			case CHK_NOTEXE :
				puts("VxDLIB Error : UNPACK : File is not an EXEcutable");
				break;

			case CHK_NOTNEWEXE :
				puts("VxDLIB Error : UNPACK : File is not a New EXEcutable");
				break;

			case CHK_NOTWEXE :
			case CHK_W3 :
				puts("VxDLIB Error : UNPACK : File is not a W4 EXEcutable");
				break;

			case CHK_W4 :
				if (access(argv[1],0x06)) // Read/Write Permission
				{
					puts("VxDLIB Error : UNPACK : File does not have Read/Write permission");
				}
				else
				{
					if (!unpack_vxd(argv[1],tmp_name))
					{
						unlink(argv[1]);
						rename(tmp_name,argv[1]);
					}
				}

				break;

			default :
				puts("VxDLIB Error : UNPACK : Unsuitable file");
		}
	}

	if (extract_option)
	{
		if (argc < 3)
		{
			puts("EXTRACT : Insufficient parameters");
			exit(1);
		}

		switch(chk_file(argv[1]))
		{
			case CHK_CANTOPEN :
				puts("VxDLIB Error : EXTRACT : Can't open file");
				break;

			case CHK_NOTEXE :
				puts("VxDLIB Error : EXTRACT : File is not an EXEcutable");
				break;

			case CHK_NOTNEWEXE :
				puts("VxDLIB Error : EXTRACT : File is not a New EXEcutable");
				break;

			case CHK_NOTWEXE :
				puts("VxDLIB Error : EXTRACT : File is not a W? EXEcutable");
				break;

			case CHK_W3 :
				extract_vxd(argv[1],argv[2]);
				break;

			case CHK_W4 :
				if (!unpack_vxd(argv[1],tmp_name))
				{
					extract_vxd(tmp_name,argv[2]);

					unlink(tmp_name);
				}

				break;

			default :
				puts("VxDLIB Error : EXTRACT : Unsuitable file");
		}
	}
}

