/* 
wispy.c -- Windows interrupt spy

Copyright (c) 1991 Andrew Schulman

from the forthcoming book UNDOCUMENTED WINDOWS, to be published by
Addison-Wesley in early 1992

syntax:  wispy [0x]intno

examples: to trap protected-mode INT 21h calls:
    wispy 0x21
        
to trap DPMI calls:
    wispy 0x31
        
to trap INT 2fh calls:
    wispy 0x2f
        
problems:
    -- can't be unloaded properly; have to exit Windows
    -- can't run multiple instances right now; should be able to run wispy
       0x31 and wispy 0x21 at same time
    -- displayed task name is sometimes stale
           
wispy 0x31 and wispy 0x2f show Windows use of undocumented DPMI calls

the undoc 0x31 calls are in old versions of DPMI spec

the undoc 0x2f calls are briefly mentioned in DDK int2fapi.inc
*/

#include "windows.h"
#include "winio.h"
#include <stdlib.h>
#include <dos.h>
#include <string.h>

#pragma pack(1)

typedef struct {
#ifdef __BORLANDC__
    unsigned bp,di,si,ds,es,dx,cx,bx,ax;
#else
    unsigned es,ds,di,si,bp,sp,bx,dx,cx,ax;     /* same as PUSHA */
#endif
    unsigned ip,cs,flags;
    } REG_PARAMS;
    
#define BUF_SIZE    1000
#define MAX_REQU    (BUF_SIZE-1)

#define SET_NEXT()  { if (next==MAX_REQU) next=0; else next++; }

typedef struct {
    int type;
    REG_PARAMS r;
    HANDLE task;
    } REQUEST;

#define REQU_FREE       'f'     /* request block not in use */
#define REQU_USE        'u'     /* in use */

#define BEGIN_CRIT_SEC()        _asm cli
#define END_CRIT_SEC()          _asm sti

#define IS_REQUEST(buf) \
    (buf[next].type != REQU_FREE)
        
#define BUFFER_FULL(buf) \
    IS_REQUEST(buf)
    
static REQUEST far *requ;

int WRT_REQUEST(REQUEST far *buf, REG_PARAMS far *pr, HANDLE task)
{
    static int next = 0;
    if (BUFFER_FULL(buf))
        return 0;
    //BEGIN_CRIT_SEC();
    _fmemcpy(&buf[next].r, pr, sizeof(REG_PARAMS));
    buf[next].type = REQU_USE;
    buf[next].task = task;
    //END_CRIT_SEC();
    SET_NEXT();
    return 1;
}

int RD_REQUEST(REQUEST far *buf, REG_PARAMS *pr, HANDLE *ptask)
{
    static int next = 0;
    if (! IS_REQUEST(buf))
        return 0;
    BEGIN_CRIT_SEC();
    _fmemcpy(pr, &buf[next].r, sizeof(REG_PARAMS));
    *ptask = buf[next].task;
    buf[next].type = REQU_FREE;
    END_CRIT_SEC();
    SET_NEXT();
    return 1;
}

REQUEST far *INIT_REQUBUF(void)
{
    REQUEST far *buf, far *p;
    int i;
    if (! (buf = GlobalLock(GlobalAlloc(GMEM_FIXED | GMEM_ZEROINIT, 
        BUF_SIZE * sizeof(REQUEST)))))
        return 0;
    for (i=BUF_SIZE, p=buf; i--; p++)
        p->type = REQU_FREE;
    return buf;
}

typedef void (_interrupt _far _export *INTRFUNC)(); 

extern void far _chain_intr(INTRFUNC);

INTRFUNC _dpmi_getvect(unsigned intno)
{
    _asm {
        mov ax, 0204h
        mov bl, byte ptr intno
        int 31h
        /* return CX:DX in DX:AX */
        mov ax, dx
        mov dx, cx
        }
}

int _dpmi_setvect(unsigned intno, INTRFUNC handler)
{
    _asm {
        mov ax, 0205h
        mov bl, byte ptr intno
        mov cx, word ptr handler+2
        mov dx, word ptr handler
        int 31h
        }
    return (_dpmi_getvect(intno) == handler);
}

void _interrupt _far _export dos(REG_PARAMS r);

static INTRFUNC old;
static unsigned long calls = 0;
static unsigned long lost = 0;

static HANDLE wispy_task;
static int intno = 0x21;    // INT 21h default 

void on_close(void)
{
    char buf[80];
    sprintf(buf, "%lu calls to INT %02xh\n (lost %lu)", 
        calls, intno, lost);
    MessageBox(NULL, buf, "WISPY", MB_OK);
    
    if (! _dpmi_setvect(intno, old))
        MessageBox(NULL, "couldn't restore INT", "WISPY", MB_OK);
    
    // big trouble here until Mike Maurice found bug: thanks, Mike!
}

unsigned verr(unsigned short sel)
{
    _asm mov ax, 1
    _asm verr word ptr sel
    _asm je short okay
    _asm dec ax
okay:;  
}

void display_int(REG_PARAMS *pr)
{
    printf("AX=%04x BX=%04x CX=%04x DX=%04x DS=%04x SI=%04x DI=%04x "
           "CS:IP=%04x:%04xh\n",
        pr->ax, pr->bx, pr->cx, pr->dx, 
        pr->ds, pr->si, pr->di, pr->cs, pr->ip);
}

void display_dos_int(REG_PARAMS *pr)
{
    int ah = pr->ax >> 8;
    printf("(%02x) ", ah);
    switch (ah)
    {
        case 0x0e: // set disk
            printf("SET DISK %02xh\n", pr->dx & 0xFF);
            break;
        case 0x19: // get current disk
            printf("GET DISK\n");
            break;
        case 0x25: // set vect
            printf("SET VECT %02xh %04x:%04x\n", 
                pr->ax & 0xFF, pr->ds, pr->dx);
            break;
        case 0x30: // get vers
            printf("GET DOS VERS\n");
            break;
        case 0x35: // get vect
            printf("GET VECT %02xh\n", pr->ax & 0xFF);
            break;
        case 0x3B: // set curr dir
            printf("CHDIR %Fs\n", MK_FP(pr->ds, pr->dx));
            break;
        case 0x3d: // open
            printf("OPEN %Fs\n", MK_FP(pr->ds, pr->dx));
            break;
        case 0x44: // IOCTL
            if ((pr->ax & 0xFF) == 0x00)    // get dev info
                printf("IOCTL GET DEV INFO %04xh\n", pr->bx);
            else
                printf("IOCTL "); display_int(pr);
            break;
        case 0x4b: // exec
            printf("EXEC %Fs\n", MK_FP(pr->ds, pr->dx));
            break;
        case 0x47: // get curr dir
            printf("GET CURR DIR\n");
            break;
        case 0x4c: // exit
            printf("EXIT %02xh\n", pr->ax & 0xff);
            break;
        case 0x4e: // find first
            printf("FIND FIRST %Fs\n", MK_FP(pr->ds, pr->dx));
            break;
        case 0x4f: // find next
            printf("FIND NEXT\n");
            break;
        case 0x55: // undoc: create PSP
            printf("(UNDOC) MK PSP %04xh size=%04xh\n", pr->dx, pr->si);
            break;
        default:
            display_int(pr);
    }
}

void display_dpmi_int(REG_PARAMS *pr)
{
    unsigned bl = pr->bx & 0xFF;
    printf("(%04x) ", pr->ax);
    switch (pr->ax)
    {
        case 0x0000: // alloc LDT descriptors
            printf("ALLOC LDT DESC n=%u\n", pr->cx);
            break;
        case 0x0001: // free LDT descriptor
            printf("FREE LDT DESC %04xh\n", pr->bx);
            break;
        case 0x0002: // seg to desc
            printf("SEG TO DESC %04xh\n", pr->bx);
            break;
        case 0x0003: // get next sel incr
            printf("GET NEXT SEL INCR\n");
            break;
        case 0x0004: // reserved: Lock Selector Memory
            printf("(RESERVED) LOCK SEL %04xh\n", pr->bx);
            break;
        case 0x0005: // reserved
            printf("(RESERVED) UNLOCK SEL %04xh\n", pr->bx);
            break;
        case 0x0009: // set descriptor access rights
            printf("SET DESC ACCRGTS %04xh %04xh\n", pr->bx, pr->cx);
            break;
        case 0x0200: // get rmode intvec
            printf("GET RMODE VEC %02xh\n", bl);
            break;
        case 0x0201: // set rmode intvec
            printf("SET RMODE VEC %02xh %04x:%04xh\n", bl, pr->cx, pr->dx);
            break;
        case 0x0202: // get except vec
            printf("GET EXCEPT VEC %02xh\n", bl);
            break;
        case 0x0203: // set except vec
            printf("SET EXCEPT VEC %02xh %04x:%04xh\n", bl, pr->cx, pr->dx);
            break;
        case 0x0204: // get pmode intvect
            printf("GET PMODE VEC %02xh\n", bl);
            break;
        case 0x0205: // set pmode intvect
            printf("SET PMODE VEC %02xh %04x:%04xh\n", bl, pr->cx, pr->dx);
            break;
        case 0x0300: // simulate rmode intr
            printf("CALL RMODE INTR %02xh\n", bl);
            break;
        case 0x0301: // call rmode proc with far ret frame
            printf("CALL RMODE PROC %04x:%04xh\n", pr->es, pr->di);
            break;
        case 0x0302: // call rmode proc with iret frame
            printf("CALL RMODE PROC (IRET) %04x:%04xh\n", pr->es, pr->di);
            break;
        case 0x0500: // get free memory information
            printf("GET FREE MEM INFO\n");
            break;
        case 0x0501: // allocate memory block
            printf("ALLOC MEM nbytes=%x%04xh\n", pr->bx, pr->cx);
            break;
        case 0x0502: // free memory block
            printf("FREE MEM %04x%04xh\n", pr->si, pr->di);
            break;
        case 0x0600: // lock linear region
            printf("LOCK LINEAR @%04x%04x nbytes=%x%04xh\n",
                pr->bx, pr->cx, pr->si, pr->di);
        case 0x0700: // reserved: Mark Paging Candidates
            printf("(RESERVED) MARK PAGING CANDIDATES @%04x%04xh "
                "npages=%x%04xh\n",
                pr->bx, pr->cx, pr->si, pr->di);
        case 0x0701: // reserved: Discard Pages
            printf("(RESERVED) DISCARD PAGES @%04x%04xh "
                "npages=%x%04xh\n",
                pr->bx, pr->cx, pr->si, pr->di);
            break;
        case 0x0703: // discard page contents
            printf("DISCARD PAGES @%04x%04x nbytes=%x%04xh\n",
                pr->bx, pr->cx, pr->si, pr->di);
            break;
        default:
            display_int(pr);
    }
}

void display_2f_int(REG_PARAMS *pr)
{
	unsigned ah = pr->ax >> 8;
	unsigned al = pr->ax & 0xFF;
    printf("(%04x) ", pr->ax);
	switch (ah)
	{
		case 0x16:	// see DDK include/int2fapi.inc
			switch (al)
			{
				case 0x00:
					printf("W386_GETVERSION\n");
					break;
				case 0x02:
					printf("W386_OLD_GET_VMID_API\n");
					break;
				case 0x05:
					printf("W386_STARTUP\n");
					break;
				case 0x06:
					printf("W386_EXIT\n");
					break;
				case 0x07:
					printf("W386_DEVICE_BROADCAST\n");
					break;
				case 0x08:
					printf("W386_STARTUP_COMPLETE\n");
					break;
				case 0x09:
					printf("W386_BEGIN_EXIT\n");
					break;
				case 0x80:
					printf("W386_RELEASE_TIME\n");
					break;
				case 0x81:
					printf("W386_BEGIN_CRITICAL\n");
					break;
				case 0x82:
					printf("W386_END_CRITICAL\n");
					break;
				case 0x83:
					printf("W386_GET_CUR_VMID\n");
					break;
				case 0x84:
					printf("W386_GET_DEVICE_API\n");
					break;
				case 0x85:
					printf("W386_SWITCH_AND_CALL\n");
					break;
				case 0x86:
					printf("W386_TEST_INT31_AVAIL\n");
					break;
				case 0x87:
					printf("W386_GET_PM_SWITCH_ADDR\n");
					break;
				case 0x88:
					printf("W386_GET_LDT_BASE_SEL\n");
					/* KRNL386.ASM: 
						mov ax, 1688h
						mov bx, 0BADh
						int 2Fh
						or ax, ax
						jnz ...
						mov LDT, bx
					*/
					break;
				case 0x89:
					printf("W386_WIN_KERNEL_IDLE\n");
					break;
				default:
					display_int(pr);
			}
			break;
		case 0x40:	// cf. DDK, pp. 2-16 - 2-18 (also OS/2 DOS box??)
			switch (al)
			{
				case 0x00:
					printf("ENABLE VM-ASSISTED SAVE/RESTORE\n");
					break;
				case 0x01:
					printf("NOTIFY BACKGROUND SWITCH\n");
					break;
				case 0x02:
					printf("NOTIFY FOREGROUND SWITCH\n");
					break;
				case 0x03:
					printf("ENTER CRITICAL SECTION\n");
					break;
				case 0x04:
					printf("EXIT CRITICAL SECTION\n");
					break;
				case 0x05:
					printf("SAVE VIDEO REGISTER STATE\n");
					break;
				case 0x06:
					printf("RESTORE VIDEO REGISTER STATE\n");
					break;
				case 0x07:
					printf("DISABLE VM-ASSISTED SAVE/RESTORE\n");
					break;
				default:
					display_int(pr);
			}
			break;
		default:
			display_int(pr);
	}
}
		
long on_time(HWND hwnd, unsigned message, WORD wParam, LONG lParam)
{
    REG_PARAMS r;
    HANDLE task, hModule;
    char modname[16];
    
//  winio_setpaint(FALSE);
    while (RD_REQUEST(requ, &r, &task))
    {
        if (task && (verr(task)))
        {
            char far *ptask = MK_FP(task, 0);
            _fmemcpy(modname, &ptask[0xf2], 8);     // undoc!! work in 3.1??
            modname[8] = '\0';
        }
        else
            sprintf(modname, "%04x?", task);

        printf("<%s> ", modname);
        if (intno == 0x21)          // DOS
            display_dos_int(&r);
        else if (intno == 0x31)     // DPMI
            display_dpmi_int(&r);
		else if (intno == 0x2f)		// INT 2Fh
			display_2f_int(&r);		
        else                        // other
            display_int(&r);        
        winio_yield();
    }
//  winio_setpaint(TRUE);
    return 0;
}

int axtoi(char *s)
{
    int ret;
    if (s[0]=='0' && s[1]=='x')
    {
        sscanf(s+2, "%x", &ret);
        return ret;
    }
    else
        return atoi(s);
}

main(int argc, char *argv[])
{
    char buf[128];
        
    if (argc > 1)
        intno = axtoi(argv[1]);
    
    sprintf(buf, "WISPY -- INT %02xh", intno);
    winio_settitle(buf);
    
    if (! (requ = INIT_REQUBUF()))
    {
        MessageBox(NULL, "INIT_REQUBUF fail!", "WISPY", MB_OK);
        return 1;
    }
    
    wispy_task = GetCurrentTask();
    
    /* hook INT XXh */
    winio_onclose(on_close);
    old = _dpmi_getvect(intno);
    if (! _dpmi_setvect(intno, dos))
    {
        printf("set INT %02xh failed!\n", intno);
        return 1;
    }

    wmhandler_set(WM_TIMER, on_time);
    if (! SetTimer(winio_hwnd(), 1, 250, NULL)) // every .25 seconds
    {
        winio_warn(FALSE, "can't create timer");
        winio_close();
    }
    
    return 0;   // go resident via WINIO yield loop
}

void _interrupt _far _export dos(REG_PARAMS r)
{
    HANDLE task = GetCurrentTask();     // seems okay from intr handler
        
    if (task != wispy_task)         // don't show my own DPMI requests
    {
        // problem: by only passing along regs, info may be stale by the
        // time that timer process sees it!!
        
        if (! WRT_REQUEST(requ, &r, task))
            lost++;     // buffer overflow
        calls++;
    }
    _chain_intr(old);
    
    // need some way to get return regs!
}
    
